import os
from glob import glob
from unittest import TestCase

from sqlalchemy import MetaData

from openmodule.config import database_folder
from openmodule.database.database import Database

_first_start = True


def delete_all_tables(database: Database):
    assert any(x in database.db_folder for x in ["test"]), "deleting all tables is only for testing"
    metadata = MetaData(bind=database._engine)
    metadata.reflect()
    with database._engine.connect() as con:
        trans = con.begin()

        for table in reversed(metadata.sorted_tables):
            con.execute(table.delete())
        trans.commit()


class SQLiteTestMixin(TestCase):
    """
    Mixin for database cleanup in test cases
    * use create_database = True for an automatic generation of a database
    * use create_database = False and set the database directly
    """
    create_database = True
    database = None
    database_folder = database_folder()
    alembic_path = "../src/database"
    database_name = "database"

    @classmethod
    def setUpClass(cls) -> None:
        # we only know which databases are in use on tear down, so truncating only works in teardown
        # but in order to not be annoyed by failed tests which left broken databases, we delete all databases
        # once initially
        global _first_start
        if _first_start:
            for file in glob(os.path.join(cls.database_folder, "*.sqlite3")):
                os.unlink(file)
                _first_start = False
        if cls.create_database:
            cls.database = Database(cls.database_folder, cls.database_name, cls.alembic_path)
        return super().setUpClass()

    def tearDown(self):
        super().tearDown()
        delete_all_tables(self.database)

    @classmethod
    def tearDownClass(cls):
        if cls.database:
            cls.database.shutdown()
        super().tearDownClass()



import glob
import unicodedata

import editdistance
import logging
import os
import yaml
from typing import List

from openmodule.models.base import OpenModuleModel

SEPARATOR_CHARS = " -_.:"
LEGACY_YAML_STRING = """
allowed: "0123456789ABCDEFGHIJKLMNPRSTUVWXYZ"
kept_separators: NONE
replacements:
  - from: "O"
    to: "0"
  - from: "Q"
    to: "0"
  - from: "Ä"
    to: "A"
  - from: "Ö"
    to: "0"
  - from: "Ü"
    to: "U"
exactitude: 0.0
"""


class MatchingConfig(OpenModuleModel):
    directory: str = "/data/schemes/schemes"
    default_scheme: str = "DEFAULT"
    default_version: int = 10
    default_country_order: List[str] = []
    legacy_in_default_country_order: bool = False
    edit_distance: int = 0
    use_alternatives: bool = False
    only_best: bool = True
    add_other_country_always: bool = True


class PlateYaml:
    def __init__(self, filename):
        self.scheme = ""
        self.version = -1
        self.exactitude = 0.0
        self.filename = filename
        self.log = logging.getLogger(f"PlateYaml<{os.path.basename(filename).replace('.', '-')}>")

        with open(filename, 'r') as f:
            try:
                self.settings = yaml.safe_load(f)
                self.valid_chars = self.settings["allowed"]
                self.exactitude = self.settings["exactitude"]
                self.kept_separators = self.settings.get("kept_separators", "NONE").upper()
                self.replacements = []
                for r in self.settings.get("replacements", []):
                    if r.get("from", "") != "" and "to" in r:
                        self.replacements.append((r.get("from"), r.get("to")))
            except:
                self.log.error("Error loading Plate Yaml %s, will be ignored", filename)
                return

        try:
            self.scheme, tmp = os.path.basename(filename).replace(".", "-").split("-")[:2]
            self.version = int(tmp)
        except Exception as e:
            logging.exception("Invalid Plate Yaml filename %s, will be ignored", filename)

    def convert(self, plate):
        plate = plate.upper()
        for r in self.replacements:
            plate = plate.replace(r[0], r[1])
        for c in set(plate):
            if c not in self.valid_chars:
                c2 = unicodedata.normalize('NFD', c).encode('ascii', 'ignore').decode("utf-8")
                plate = plate.replace(c, c2 if c2 in self.valid_chars else "")

        if self.kept_separators != "ALL":
            plate_stripped = plate
            for c in list(SEPARATOR_CHARS):
                plate_stripped = plate_stripped.replace(c, "")
            if self.kept_separators == "FIRST":
                for i, c in enumerate(list(plate)):
                    if c in SEPARATOR_CHARS:
                        plate = plate[:i + 1] + plate_stripped[i:]
                        break
            elif self.kept_separators == "LAST":
                for i, c in enumerate(reversed(list(plate))):
                    if c in SEPARATOR_CHARS:
                        plate = plate_stripped[:-i] + plate[-1 - i:]
                        break
            else:
                plate = plate_stripped

        return plate


class PlateYamlSelector(object):
    @classmethod
    def from_config(cls, config: MatchingConfig):
        revision_path = os.path.join(os.path.dirname(config.directory), "revision")
        return cls(
            yaml_dir=config.directory,
            revision_file=revision_path,
            default_country_order=config.default_country_order,
            legacy_in_default_country_order=config.legacy_in_default_country_order,
            always_add_other_country_to_plate_for_matching=config.add_other_country_always
        )

    def __init__(self, yaml_dir, revision_file="", default_country_order=None, legacy_in_default_country_order=False,
                 always_add_other_country_to_plate_for_matching=True):
        self.yaml_dir = yaml_dir
        self.default_country_order = default_country_order or []
        self.legacy_in_default_country_order = legacy_in_default_country_order
        self.always_add_other_country_to_plate_for_matching = always_add_other_country_to_plate_for_matching
        self.schemes = {}
        self.revision_file = revision_file
        self.revision = ""
        self.reload_yamls()

    def reload_yamls(self):
        if self.revision_file != "" and os.path.exists(self.revision_file):
            with open(self.revision_file) as f:
                new_revision = f.read()
        else:
            new_revision = "LOADED"

        if new_revision != self.revision:
            yaml_filenames = glob.glob(os.path.join(self.yaml_dir, "*.yml"))
            self.schemes.clear()
            for fn in yaml_filenames:
                plate_yaml = PlateYaml(fn)
                if plate_yaml.version >= 0:  # check, if loading was successful
                    if plate_yaml.scheme in self.schemes:
                        self.schemes[plate_yaml.scheme][plate_yaml.version] = plate_yaml
                    else:
                        self.schemes[plate_yaml.scheme] = {plate_yaml.version: plate_yaml}
            if self.schemes.get("LEGACY", {}).get(0, None) is None:
                with open("/tmp/LEGACY-0.yml", 'w') as f:
                    f.write(LEGACY_YAML_STRING)
                legacy_yaml = PlateYaml("/tmp/LEGACY-0.yml")
                if legacy_yaml.version < 0:
                    raise Exception("NO LEGACY YAML COULD BE LOADED / CREATED")
            else:
                legacy_yaml = self.schemes["LEGACY"][0]
                del self.schemes["LEGACY"]
            if "DEFAULT" not in self.schemes:
                self.schemes["DEFAULT"] = {}

            for key in self.schemes.keys():
                self.schemes[key][0] = legacy_yaml

            self.revision = new_revision

    def get_all_exactitudes_decreasing(self):
        exactitudes = set([])
        for s in self.schemes.keys():
            for v in self.schemes[s].keys():
                exactitudes.add(self.schemes[s][v].exactitude)
        return sorted(exactitudes, reverse=True)

    def get_best_yaml(self, scheme, version):
        if scheme not in self.schemes:
            scheme = "DEFAULT"

        versions = sorted(self.schemes.get(scheme, {}).keys(), reverse=True)
        for v in versions:
            if v <= version:
                return self.schemes[scheme][v]

        assert False, "NO YAML FOUND! THIS SHOULD NOT HAPPEN!"

    def convert(self, plate, scheme, version, return_matched_yaml=False):
        if return_matched_yaml:
            best_yaml = self.get_best_yaml(scheme, version)
            return {"plate": best_yaml.convert(plate), "scheme": best_yaml.scheme, "version": best_yaml.version,
                    "exactitude": best_yaml.exactitude}
        else:
            return self.get_best_yaml(scheme, version).convert(plate)

    def create_db_record(self, plate, country, scheme, version, id=None, **kwargs):
        """
        This function generates a valid database record.
        :param plate: read plate
        :param country: read country
        :param scheme: scheme that should be used
        :param version: version that should be used
        :param id: id only used in testcases
        :param kwargs: additional arguments, e.g. skidata hostcom provides ticket id
        :return: valid database entry
        """
        converted = self.convert(plate, scheme, version, return_matched_yaml=True)
        payload = {
            "plate": plate,
            "country": country,
            "plate_scheme": scheme,
            "plate_scheme_version": version,
            "plate_conv": converted.get("plate", plate),
            "exactitude": converted.get("exactitude", 0.0),
            "used_scheme": converted.get("scheme", scheme),
            "used_version": converted.get("version", version),
            "id": id,
        }
        if kwargs:
            if any(x in payload.keys() for x in kwargs.keys()):
                raise Exception("additional arguments have key that is part of the matching db record!")
            payload.update(kwargs)
        return payload

    def create_plate_for_matching(self, plate, country="OTHER", plate_confidence=0.5, country_confidence=0.5,
                                  alternatives=None):
        """
        This functions generates the plate payload that is needed by the `match_plate` function.
        :param plate: read plate
        :param country: read country
        :param plate_confidence: confidence of the plate
        :param country_confidence: confidence of the read country
        :param alternatives: plate alternatives
        :return: valid plate for `match_plate`
        """
        if alternatives is None:
            alternatives = []
        country = [{"code": country, "confidence": country_confidence}]
        if self.always_add_other_country_to_plate_for_matching and country != "OTHER":
            country.append({"code": "OTHER", "confidence": 0.5})
        return {
            "plate": {"plate": plate, "confidence": plate_confidence},
            "country": country,
            "alternatives": alternatives
        }


    def insert_into_db(self, db, record):
        """
        This function inserts a record correctly into the db.
        :param db: database to insert
        :param record: db record created with `create_db_record`
        """
        country = record["country"]
        if country not in db:
            db[country] = {}
        if record["used_version"] not in db[country]:
            db[country][record["used_version"]] = []
        db[country][record["used_version"]].append(record)

    def insert_into_db_and_create_record(self, db, plate, country, scheme, version, id=None, **kwargs):
        """
        This function adds a record correctly into db with create_db_record and then inserting the record
        :param db: database to insert
        :param plate: read plate
        :param country: read country
        :param scheme: scheme that should be used
        :param version: version that should be used
        :param id: id only used in testcases
        """
        record = self.create_db_record(plate, country, scheme, version, id, **kwargs)
        self.insert_into_db(db, record)

    def check_ed0(self, db_cv, plate, matches, num_alt, return_only_best=True):
        if db_cv and "used_scheme" in db_cv[0]:
            p_conv = self.convert(plate, db_cv[0]["used_scheme"], db_cv[0]["used_version"])
            for db_entry in db_cv:
                if p_conv == db_entry["plate_conv"]:
                    matches[num_alt].append(db_entry)
                    if return_only_best:
                        return db_entry, matches
        return None, matches

    def check_edx(self, db_cv, plate, max_ed, matches, return_only_best=True):
        if db_cv and "used_scheme" in db_cv[0]:
            p_conv = self.convert(plate, db_cv[0]["used_scheme"], db_cv[0]["used_version"])
            for db_entry in db_cv:
                ed = editdistance.eval(p_conv, db_entry["plate_conv"])
                if ed <= max_ed:
                    matches[ed].append(db_entry)
                if ed == 0 and return_only_best:
                    return db_entry, matches
        return None, matches

    def match_plate(self, plate, db, use_alternatives=True, max_ed=0, return_only_best=True):
        """
        Expected format of plate: {"plate": {"plate": <plate>, ...},
                                       "country": [{"country": <A/D/...>, "confidence": <0.0-1.0>}],
                                       "alternatives": [{"plate": <alt_plate1>, "confidence": <0.0-1.0>},...]}
        Expected format of db: {<Country1>: {<Version1>: [<plate1>, <plate2>,...]}}
            plates in converted form (converted using the best matching yaml, sorted into convert scheme and version,
                                                                                not db entry scheme and version)
        ED0 if use_alternatives (ed1 and alternatives does not make much sense)
        if return_only_best, matching will be stopped if ED0 match was found
        """
        self.reload_yamls()

        if use_alternatives:
            max_ed = 0
            plates = plate.get("alternatives", [])
            if not plates:
                logging.warning("use alternatives is active but no alternatives present! falling back to plate")
                plates = [plate.get("plate", {})]
            matches = [[]] * len(plates)
        else:
            # If use_alternatives is false then match current plate with edX like the controller did
            plates = [plate.get("plate", {})]
            matches = [[]] * (max_ed + 1)

        for num_alt, p in enumerate(plates):
            check_other = False
            # first check based on countries given in plate
            checked_countries = []
            for c in plate.get("country", []):
                country = c["code"]
                if country == "OTHER":
                    check_other = True
                    break
                if country not in db:
                    continue
                for version in sorted(db[country].keys(), reverse=True):
                    if max_ed == 0:
                        res, matches = self.check_ed0(db[country].get(version, []), p["plate"], matches, num_alt,
                                                      return_only_best)
                    else:
                        res, matches = self.check_edx(db[country].get(version, []), p["plate"], max_ed, matches,
                                                      return_only_best)
                    if res is not None:
                        return [[res]]
                checked_countries.append(country)

            # if other specified, check other countries
            if check_other:
                no_legacy_checked_countries = []
                # first check in order given in config, but ignoring legacy (if not legacy_in_default_country_order)
                for country in self.default_country_order:
                    if country in checked_countries or country not in db:
                        continue
                    versions = sorted(db[country].keys(), reverse=True)
                    if not self.legacy_in_default_country_order and 0 in versions:
                        versions.remove(0)
                    for version in versions:
                        if max_ed == 0:
                            res, matches = self.check_ed0(db[country].get(version, []), p["plate"], matches, num_alt,
                                                          return_only_best)
                        else:
                            res, matches = self.check_edx(db[country].get(version, []), p["plate"], max_ed, matches,
                                                          return_only_best)
                        if res is not None:
                            return [[res]]
                    if self.legacy_in_default_country_order:
                        checked_countries.append(country)
                    else:
                        no_legacy_checked_countries.append(country)

                # rest is checked in decreasing version order
                for exactitude in self.get_all_exactitudes_decreasing():
                    for country, db_country in db.items():
                        if country in checked_countries or exactitude != 0 and country in no_legacy_checked_countries:
                            continue
                        for version, db_cv in db_country.items():
                            if db_cv and db_cv[0]["exactitude"] == exactitude:
                                if max_ed == 0:
                                    res, matches = self.check_ed0(db[country].get(version, []), p["plate"], matches,
                                                                  num_alt, return_only_best)
                                else:
                                    res, matches = self.check_edx(db[country].get(version, []), p["plate"], max_ed,
                                                                  matches, return_only_best)
                                if res is not None:
                                    return [[res]]

            if return_only_best:
                for edm in matches:
                    if len(edm) > 0:
                        return [[edm[0]]]

        return matches



from datetime import datetime
from typing import Optional

from pydantic import Field

from openmodule.models.base import ZMQMessage, OpenModuleModel, Gateway
from openmodule.models.vehicle import Medium, LPRMedium


class PresenceMedia(OpenModuleModel):
    lpr: Optional[LPRMedium]
    qr: Optional[Medium]
    nfc: Optional[Medium]
    pin: Optional[Medium]


class PresenceBaseMessage(ZMQMessage):
    vehicle_id: int
    unsure: bool = False
    source: str
    present_area_name: str = Field(..., alias="present-area-name")
    last_update: datetime
    gateway: Gateway
    medium: PresenceMedia


class PresenceBackwardMessage(PresenceBaseMessage):
    type: str = "backward"
    leave_time: str = Field(..., alias="leave-time")


class PresenceForwardMessage(PresenceBaseMessage):
    type: str = "forward"
    leave_time: str = Field(..., alias="leave-time")


class PresenceLeaveMessage(PresenceBaseMessage):
    type: str = "leave"


class PresenceEnterMessage(PresenceBaseMessage):
    type: str = "enter"


class PresenceChangeMessage(PresenceBaseMessage):
    type: str = "change"


import uuid

from sqlalchemy import Column, String, Integer
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class DatabaseTestModel(Base):
    __tablename__ = "test"
    id = Column(String, default=lambda: str(uuid.uuid4()), unique=True, primary_key=True)
    value1 = Column(Integer, default=1)
    value2 = Column(Integer, default=2)
    string = Column(String, default="initial")



from unittest import TestCase

from openmodule.models.vehicle import Vehicle


class VehicleTestCase(TestCase):
    def test_vehicle_str(self):
        # just checks that the relevant info is
        vehicle = Vehicle(id=1613386532124866889,
                          qr={"id": "someverylongqrcodeitisreallylongohyeah", "type": "qr"},
                          lpr={"id": "test", "type": "lpr", "country": {"code": "D"}})

        vehicle_str = str(vehicle)

        self.assertIn("id:4866889", vehicle_str)  # only the last 7 digits are printed because they are random enough
        self.assertIn("somevery", vehicle_str)
        self.assertIn("...", vehicle_str)  # long medium id's are truncated
        self.assertIn("test", vehicle_str)



import threading
import time
from unittest import TestCase

import zmq

from openmodule_test.zeromq import patch_bind_string, TestBroker, TestClient, fake_config, ZMQTestMixin


class ZMQUtilsTest(TestCase):
    def test_patch_zmq_bind_str(self):
        self.assertEqual("tcp://127.0.0.1:10100", patch_bind_string("tcp://*:10100"))
        self.assertEqual("tcp://0.0.0.0:10100", patch_bind_string("tcp://0.0.0.0:10100"))
        self.assertEqual("unix:///tmp/asdf", patch_bind_string("unix:///tmp/asdf"))
        self.assertEqual("inproc://endpoint", patch_bind_string("inproc://endpoint"))

    def test_fake_config(self):
        broker = TestBroker("inproc://sub", "inproc://pub")
        config = fake_config(broker)

        self.assertEqual("inproc://pub", config.BROKER_PUB)
        self.assertEqual("inproc://sub", config.BROKER_SUB)


class TestBrokerTest(TestCase):
    def test_start_and_stop(self):
        broker = TestBroker("tcp://127.0.0.1:10001", "tcp://127.0.0.1:10002")
        broker.start()
        broker.stop()

    def _run_message_test(self, sub, pub):
        broker = TestBroker(sub, pub)
        broker.start()

        pub_socket = broker.context.socket(zmq.PUB)
        sub_socket = broker.context.socket(zmq.SUB)

        try:
            pub_socket.connect(sub)
            sub_socket.connect(pub)
            sub_socket.subscribe(b"test-topic")

            # sleep to prevent slow joiner
            time.sleep(1)

            pub_socket.send_multipart((b"test-topic", b"message"))
            topic, message = sub_socket.recv_multipart()

            self.assertEqual(b"test-topic", topic)
            self.assertEqual(b"message", message)

        finally:
            pub_socket.close()
            sub_socket.close()
            broker.stop()

    def test_single_message_tcp(self):
        self._run_message_test(
            "tcp://127.0.0.1:10003",
            "tcp://127.0.0.1:10004",
        )

    def test_single_message_inproc(self):
        self._run_message_test(
            "inproc://sub",
            "inproc://pub",
        )


class TestClientTest(TestCase):
    def test_detect_infinite_linger(self):
        broker = TestBroker("tcp://127.0.0.1:10099", "tcp://127.0.0.1:10098")
        broker.start()

        socket = broker.context.socket(zmq.PUB)
        socket.connect("tcp://127.0.0.1:10098")

        with self.assertRaises(AssertionError) as e:
            broker.stop()
        self.assertIn("linger", str(e.exception).lower())

    def _run_test_start_and_stop(self, sub, pub):
        broker = TestBroker(sub, pub)
        broker.start()

        client = TestClient(broker)
        client.start()

        broker.stop()
        client.stop()

    def test_start_and_stop_tcp(self):
        self._run_test_start_and_stop(
            "tcp://127.0.0.1:10007",
            "tcp://127.0.0.1:10008",
        )

    def test_start_and_stop_inproc(self):
        self._run_test_start_and_stop(
            "inproc://sub",
            "inproc://pub",
        )

    def test_recv_timeout_during_other_messages(self):
        # test client has two timeout conditions
        # - no message received at all during <timeout>
        # - some messages received during <timeout> but not the one we want
        # this case tests the second one
        broker = TestBroker("tcp://127.0.0.1:10009", "tcp://127.0.0.1:10010")
        broker.start()

        client = TestClient(broker)
        client.start()

        try:
            client.subscribe(b"test1", b"test2")

            def spam():
                for x in range(15):
                    client.send(b"test1", type="some-type", data="yes!")
                    time.sleep(0.1)

            spammer = threading.Thread(target=spam)
            spammer.start()

            with self.assertRaises(TimeoutError):
                client.wait_for_message_on_topic(b"test2", timeout=0.5)

            spammer.join()

        finally:
            client.stop()
            broker.stop()

    def test_one_client_in_two_threads(self):
        broker = TestBroker("tcp://127.0.0.1:10011", "tcp://127.0.0.1:10012")
        broker.start()

        client = TestClient(broker)
        client.start()
        client.subscribe(b"test")

        try:
            counter_lock = threading.Lock()
            assertions = 0

            def second_thread_func():
                nonlocal assertions
                try:
                    client.wait_for_message_on_topic(b"test")
                except AssertionError as e:  # pragma: no cover
                    self.assertIn("the test client is not thread safe", str(e))
                    with counter_lock:
                        assertions += 1
                except TimeoutError:  # pragma: no cover
                    pass

            second_thread = threading.Thread(target=second_thread_func)
            second_thread.start()

            try:
                client.wait_for_message_on_topic(b"test")
            except AssertionError as e:  # pragma: no cover
                self.assertIn("the test client is not thread safe", str(e))
                with counter_lock:
                    assertions += 1
            except TimeoutError:  # pragma: no cover
                pass

            second_thread.join()

            self.assertEqual(1, assertions)

        finally:
            client.stop()
            broker.stop()

    def test_two_clients(self):
        # two clients should be able to start and stop independently
        broker = TestBroker("tcp://127.0.0.1:10013", "tcp://127.0.0.1:10014")
        broker.start()

        client1 = TestClient(broker)
        client1.start()

        client2 = TestClient(broker)
        client2.start()

        try:
            client1.stop()

            client2.subscribe(b"test")
            client2.send(b"test", type="some-type", data="yes!")
            message = client2.wait_for_message_on_topic(b"test")
            self.assertEqual("yes!", message["data"])
        finally:
            client2.stop()
            broker.stop()

    def test_recv_message(self):
        broker = TestBroker("tcp://127.0.0.1:10015", "tcp://127.0.0.1:10016")
        broker.start()

        client = TestClient(broker)
        client.start()

        try:
            # this assertion catches developer errors, we cannot receive on a not-subscribed topic
            with self.assertRaises(AssertionError) as e:
                client.wait_for_message_on_topic(b"test")
            self.assertIn("please subscribe", str(e.exception))

            # receive a message
            client.subscribe(b"test")
            client.send(b"test", type="some-type", data="yes!")
            message = client.wait_for_message_on_topic(b"test")
            self.assertEqual("yes!", message["data"])

            # receive a second time -> timeout
            with self.assertRaises(TimeoutError):
                client.wait_for_message_on_topic(b"test")

        finally:
            client.stop()
            broker.stop()


class ZMQTestMixinTCPTest(ZMQTestMixin, TestCase):
    topics = ["test"]
    protocol = "tcp://"

    def test_send_and_receive(self):
        self.assertTrue(self.zmq_config().BROKER_PUB.startswith("tcp://"))
        self.assertTrue(self.zmq_config().BROKER_SUB.startswith("tcp://"))

        self.zmq_client.send(b"test", type="some-type", data="yes!")
        message = self.zmq_client.wait_for_message_on_topic(b"test")
        self.assertEqual("yes!", message["data"])

    def test_client_ignores_ping_commands(self):
        # we do not want to accidentially receive ping commands used during connecting in our tests
        self.zmq_client.send(b"test", type="some-type", data="yes!")
        self.zmq_client.wait_for_message_on_topic(b"test")

        self.zmq_client._zmq_cmd("some-command", topic=b"test")
        with self.assertRaises(TimeoutError):
            self.zmq_client.wait_for_message_on_topic(b"test")


class ZMQTestMixinTest(ZMQTestMixin, TestCase):
    topics = ["test"]
    rpc_channels = ["channel"]

    def test_send_and_receive(self):
        self.zmq_client.send(b"test", type="some-type", data="yes!")
        message = self.zmq_client.wait_for_message_on_topic(b"test")
        self.assertEqual("yes!", message["data"])

    def test_rpc(self):
        rpc_server_client = TestClient(self.zmq_broker)
        rpc_server_client.start()
        rpc_server_client.subscribe(b"rpc-req-channel")

        def fake_rpc_server():
            message = rpc_server_client.wait_for_message_on_topic(b"rpc-req-channel")
            rpc_server_client.send(b"rpc-rep-channel", type="type", response={"test": "yes"}, rpc_id=message["rpc_id"])

        rpc_server = threading.Thread(target=fake_rpc_server)
        rpc_server.start()

        self.rpc("channel", "type", {"some-request": True})

        rpc_server.join()
        rpc_server_client.stop()



import time
from unittest import TestCase

from openmodule.core import init_openmodule, shutdown_openmodule
from openmodule.models.base import Direction
from openmodule.utils.presence import PresenceListener
from openmodule_test.eventlistener import MockEvent
from openmodule_test.presence import PresenceSimulator
from openmodule_test.zeromq import ZMQTestMixin


class BasePresenceTest(ZMQTestMixin, TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.core = init_openmodule(self.zmq_config(), context=self.zmq_context())
        self.presence_sim = PresenceSimulator("gate_in", Direction.IN, lambda x: self.zmq_client.send(b"presence", x))

        self.presence = PresenceListener(self.core.messages)
        self.on_enter = MockEvent()
        self.on_forward = MockEvent()
        self.on_backward = MockEvent()
        self.on_change = MockEvent()
        self.on_leave = MockEvent()
        self.presence.on_enter.append(self.on_enter)
        self.presence.on_forward.append(self.on_forward)
        self.presence.on_backward.append(self.on_backward)
        self.presence.on_change.append(self.on_change)
        self.presence.on_leave.append(self.on_leave)
        self.wait_for_dispatcher(self.core.messages)

    def tearDown(self):
        shutdown_openmodule()
        super().tearDown()


class PresenceTest(BasePresenceTest):
    def test_cannot_use_present_vehicle_without_gate(self):
        with self.assertRaises(AssertionError) as e:
            self.presence.present_vehicle.json()
        self.assertIn("please access the present vehicle per gate", str(e.exception))

        self.presence.gate = "test"
        val = self.presence.present_vehicle
        self.assertIsNone(val)

    def test_double_entry(self):
        self.presence_sim.enter(self.presence_sim.vehicle().lpr("A", "G ARIVO1"))
        self.on_enter.wait_for_call()
        self.assertEqual("G ARIVO1", self.presence.present_vehicles["gate_in"].lpr.id)

        MockEvent.reset_all_call_counts()
        self.presence_sim.current_present = None
        with self.assertLogs() as cm:
            self.presence_sim.enter(self.presence_sim.vehicle().lpr("A", "G ARIVO2"))
            self.on_enter.wait_for_call()
            self.on_leave.wait_for_call()
        self.assertEqual("G ARIVO2", self.presence.present_vehicles["gate_in"].lpr.id)
        self.assertIn("A leave will be faked", str(cm))

    def test_faulty_exit(self):
        with self.assertLogs() as cm:
            self.presence_sim.current_present = self.presence_sim.vehicle().lpr("A", "G ARIVO1")
            self.presence_sim.leave()
            time.sleep(1)
            self.assertEqual({}, self.presence.present_vehicles)
        self.assertIn("The leave will be ignored", str(cm))

        self.presence_sim.enter(self.presence_sim.vehicle().lpr("A", "G ARIVO1"))
        self.on_enter.wait_for_call()
        self.assertEqual("G ARIVO1", self.presence.present_vehicles["gate_in"].lpr.id)

        self.presence_sim.current_present = self.presence_sim.vehicle().lpr("A", "G ARIVO2")
        with self.assertLogs() as cm:
            self.presence_sim.leave()
            self.on_leave.wait_for_call()
            self.assertEqual({}, self.presence.present_vehicles)
        self.assertIn("We are fake-leaving the currently present vehicle", str(cm))

    def test_normal(self):
        self.presence_sim.enter(self.presence_sim.vehicle().lpr("A", "G ARIVO1"))
        self.on_enter.wait_for_call()
        self.assertEqual("G ARIVO1", self.presence.present_vehicles["gate_in"].lpr.id)

        self.presence_sim.leave()
        self.on_leave.wait_for_call()
        self.assertEqual({}, self.presence.present_vehicles)

    def test_other_calls(self):
        vehicle = self.presence_sim.vehicle().lpr("A", "G ARIVO1")

        self.presence_sim.forward(vehicle)
        self.on_forward.wait_for_call()

        self.presence_sim.backward(vehicle)
        self.on_backward.wait_for_call()

        self.presence_sim.change(vehicle)
        self.on_change.wait_for_call()


class GateFilterTest(BasePresenceTest):
    def setUp(self) -> None:
        super().setUp()
        self.presence2 = PresenceListener(self.core.messages, "gate_in")
        self.on_enter2 = MockEvent()
        self.presence2.on_enter.append(self.on_enter2)

    def tearDown(self):
        super().tearDown()

    def test_gate_filter(self):
        self.presence_sim.enter(self.presence_sim.vehicle().lpr("A", "G ARIVO1"))

        self.on_enter.wait_for_call()
        self.assertIn("gate_in", self.presence.present_vehicles.keys())
        self.presence.present_vehicles = {}
        self.on_enter2.wait_for_call()
        self.assertIn("gate_in", self.presence2.present_vehicles.keys())
        self.presence2.present_vehicles = {}

        MockEvent.reset_all_call_counts()

        sim_out = PresenceSimulator("gate_out", Direction.OUT, lambda x: self.zmq_client.send(b"presence", x))
        sim_out.enter(sim_out.vehicle().lpr("A", "G ARIVO2"))
        self.on_enter.wait_for_call()

        self.assertIn("gate_out", self.presence.present_vehicles.keys())
        with self.assertRaises(AssertionError):
            self.on_enter2.wait_for_call()

        self.assertEqual({}, self.presence2.present_vehicles)



from typing import Optional, Dict, Any
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.transport import Transport

from openmodule.models.base import ZMQMessage

if TYPE_CHECKING:  # pragma: no cover
    from openmodule.core import OpenModuleCore


class SentryEvent(ZMQMessage):
    type: str = "sentry"
    event: Any


class ZeroMQTransport(Transport):
    def __init__(self, core, options: Optional[Dict[str, Any]] = None):
        self.core = core
        super(ZeroMQTransport, self).__init__(options)

    def capture_event(self, event):
        message = SentryEvent(name=self.core.config.NAME, event=event)
        self.core.publish(message, b"sentry")

    def capture_envelope(self, envelope):
        raise NotImplementedError()

    def flush(self, timeout, callback=None):
        pass

    def kill(self):
        pass


def init_sentry(core: 'OpenModuleCore', extras=None, **kwargs):
    """
    This function initializes Sentry with our predefined values. This function also check if Sentry should be
    initialized.

    :param core: openmodule core instance
    :param extras: global extras that should be added to message
    :param kwargs: client supported **kwargs see ClientOptions
    """
    if should_activate_sentry(core.config):
        zmq_transport = ZeroMQTransport(core)
        environment = environment_from_config(core.config)
        sentry_sdk.init(
            dsn=None, release=core.config.VERSION, server_name=core.config.RESOURCE, environment=environment,
            transport=zmq_transport, **kwargs
        )

        extras = extras or {}
        extras.update(extra_from_config(core.config))
        with sentry_sdk.configure_scope() as scope:
            for key, value in extras.items():
                scope.set_extra(key, value)
    else:
        core.log.info("not activating sentry in debug or test mode")


def should_activate_sentry(config) -> bool:
    """
    This function checks if for the given config Sentry should be activated or not.

    :param config: current configuration
    :return: bool
    """
    if config.DEBUG or config.TESTING:
        return False
    return True


def environment_from_config(config) -> str:
    """
    This functions returns either the environment of the current configuration.

    :param config: current configuration
    :return: 'staging' or 'production'
    """
    if hasattr(config, "DEVICE_HOST") and ("test" in config.DEVICE_HOST):
        return "staging"
    else:
        return "production"


def extra_from_config(config):
    extra = {
        "name": config.NAME,
    }
    if hasattr(config, "GATE"):
        extra["gate"] = config.GATE
    return extra



from enum import Enum
from urllib.parse import urljoin

import requests
from requests.auth import HTTPBasicAuth, HTTPDigestAuth


class ApiException(Exception):
    def __init__(self, status_code, msg=""):
        if status_code in [503, 412, 500, 408]:
            retry = True
        else:
            retry = False
        self.status = status_code
        self.msg = msg
        self.retry = retry

    def __str__(self):
        return f"Api Error {self.status}: {self.msg}"


class AuthMethod(str, Enum):
    digest = "digest"
    basic = "basic"
    none = "none"


class Api:
    """
    Api template class
    provides basic request functionality
    """

    def __init__(self, host, auth_type: AuthMethod, verify_ssl: bool, user: str = "user", password: str = "password",
                 timeout=10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.server_host = host
        self.auth = self.get_auth(auth_type, user, password)
        self.verify_ssl = verify_ssl
        self.request_timeout = timeout

    @classmethod
    def get_auth(cls, auth_type, user, password):
        if auth_type == AuthMethod.basic:
            return HTTPBasicAuth(user, password)
        elif auth_type == AuthMethod.digest:
            return HTTPDigestAuth(user, password)
        else:
            return None

    @classmethod
    def _common_headers(cls):
        return {}

    def _request(self, method, url, payload=None, json_response=True, **kwargs):
        url = urljoin(self.server_host, url)
        timeout = kwargs.pop("timeout", self.request_timeout)
        headers = self._common_headers()
        headers.update(kwargs.pop("headers", {}))
        try:
            result = requests.request(method.upper(), url, **kwargs, timeout=timeout, headers=headers,
                                      verify=self.verify_ssl, json=payload, auth=self.auth)
            if 200 <= result.status_code < 300:
                if json_response:
                    return result.json()
                return result.content
        except requests.Timeout:
            raise ApiException(status_code=408, msg=f"Timeout after {timeout}s")
        except Exception as e:
            raise ApiException(status_code=400, msg=str(e))
        raise ApiException(status_code=result.status_code, msg="Server error")

    def get(self, url, payload=None, json_response=True, **kwargs):
        return self._request("GET", url, payload, json_response, **kwargs)

    def post(self, url, payload=None, json_response=True, **kwargs):
        return self._request("POST", url, payload, json_response, **kwargs)

    def put(self, url, payload=None, json_response=True, **kwargs):
        return self._request("PUT", url, payload, json_response, **kwargs)

    def delete(self, url, payload=None, json_response=True, **kwargs):
        return self._request("DELETE", url, payload, json_response, **kwargs)



#!/usr/bin/env python

# This will try to import setuptools. If not here, it fails with a message
try:
    from setuptools import setup
except ImportError:
    raise ImportError("This module could not be installed, probably because"
                      " setuptools is not installed on this computer."
                      "\nInstall ez_setup ([sudo] pip install ez_setup) and try again.")

setup(
    setup_requires=['pbr'],
    pbr=True,
    package_dir={"openmodule_test": "./"},
    packages=["openmodule_test"]
)



import time

from openmodule.utils.backend import Backend
from openmodule.models.backend import AccessRequest, CountMessage, Access
from openmodule.core import core
from openmodule.models.base import Gateway
from openmodule_test.core import OpenModuleCoreTestMixin
from openmodule_test.eventlistener import MockEvent


class TestBackend(Backend):
    __test__ = False  # otherwise pytest thinks this is a testcase

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.message_processed = MockEvent()

    checked_in = 0
    accessed = []

    def wait_for_message_process(self):
        """
        waits for at least one message to be processed
        """
        self.message_processed.wait_for_call()

    def check_in(self, message: CountMessage):
        if message.gateway.gate == "error":
            self.message_processed()
            raise Exception("test_exception")
        self.checked_in += 1
        self.message_processed()

    def check_out(self, message: CountMessage):
        if message.gateway.gate == "error":
            self.message_processed()
            raise Exception("test_exception")
        self.checked_in = self.checked_in - 1 if self.checked_in > 0 else 0
        self.message_processed()

    def check_access(self, request: AccessRequest):
        self.accessed += [request.id]
        if request.gateway.gate == "error":
            raise Exception("test_exception")
        if request.gateway.gate == "empty":
            return []
        return [Access(category="permanent-employee", start=time.time() - 1000, user="user")]


class BackendTestMixin(OpenModuleCoreTestMixin):
    """
    BackendTestMixin with helper functions for testing backends
    * set the backend_class
    """
    backend_class = None

    @classmethod
    def setUpClass(cls) -> None:
        assert cls.backend_class, "set a backend_class"
        return super().setUpClass()

    def setUp(self):
        super().setUp()
        self.backend = self.backend_class(core())
        self.transaction_count = 0
        self.messages_by_transaction = {}
        self.zone_count = {}

    def tearDown(self):
        self.backend.shutdown()
        super().tearDown()

    @classmethod
    def create_access_request(cls, gate="gate1", direction="in", medium_type="lpr", id="GARIVO1") -> AccessRequest:
        """ creates standard count message
        :param gate: gate
        :param direction: direction
        :param medium_type: medium type (Medium)
        :param id: id of medium

        :return: access request (AccessRequest)
        """
        gateway = Gateway(gate=gate, direction=direction)
        return AccessRequest(gateway=gateway, name="backend", medium_type=medium_type, id=id)

    def check_auth(self, **kwargs):
        """
         standard check_auth call
        : param kwargs: Parameters to overwrite in the AccessRequest -> create_access_request
        """
        assert self.backend, "you need to set a backend"
        request = self.create_access_request(**kwargs)
        return self.backend.rpc_check_access(request, None)

    def create_count_message(self, transaction_id=None, *, resource="arivotest", user="api-backend_user", gate="gate1",
                             direction="in", medium="lpr",
                             medium_id="GARIVO1", zone="zone1", category="permanent-employee", correction=False,
                             error=None, **kwargs):
        """ creates standard count message
        :param transaction_id: Transaction ID for multiple sending of the same message, autogenerated if None
        :param resource: resource
        :param user: user
        :param gate: gate
        :param direction: direction
        :param medium: medium type (Medium)
        :param medium_id: id of medium
        :param zone: zone
        :param category: category (Category)
        :param correction: Flag if the message was only a correction
        :param error: Error if present, should not be used directly
        :param kwargs: Additional parameters for errors, should not be used directly
        :return: transaction_id
        """
        gateway = Gateway(gate=gate, direction=direction)
        if transaction_id is None:
            transaction_id = f"transaction-{self.transaction_count}"
            self.transaction_count += 1

        data = kwargs
        data.update(dict(user=user, gateway=gateway, medium_type=medium, id=medium_id, category=category,
                         real=not correction, transaction_id=transaction_id, count=0, error=error,
                         zone=zone, name="count", type="count", timestamp=time.time()), resource=resource)
        message = CountMessage(**data)
        self.messages_by_transaction[transaction_id] = message
        return transaction_id

    def check_in(self, transaction_id=None, send=True, **kwargs) -> str:
        """ standard check_in call
        :param transaction_id: Transaction ID of created count message
        :param send: Pass the check in call to the backend
        :param kwargs: message kwargs if transaction_id is None
        :return: transaction_id
        """

        assert self.backend, "you need to set a backend"
        if transaction_id is None:
            transaction_id = self.create_count_message(**kwargs)
        message = self.messages_by_transaction[transaction_id]

        self.zone_count[message.zone] = 1 + self.zone_count.get(message.zone, 0)
        message.count = self.zone_count[message.zone]
        message.gateway.direction = "in"
        if send:
            self.backend.check_in(message)
        return transaction_id

    def check_in_double_entry(self, transaction_id) -> str:
        """ check_in call with double entry
        :param transaction_id: previous entry transaction
        """

        old_message = self.messages_by_transaction[transaction_id]
        assert old_message.gateway.direction == "in", "needs previous check_in message"
        data = {k: v for k, v in old_message.dict().items() if k != "transaction_id"}
        data["error"] = "double_entry"
        data["previous_transaction_id"] = [transaction_id]
        new_transaction_id = self.create_count_message(**data)
        self.check_in(new_transaction_id)
        return new_transaction_id

    def check_out(self, transaction_id) -> CountMessage:
        """ standard check_out call
        :param transaction_id: Transaction ID of the check_in message
        """
        assert self.backend, "you need to set a backend"
        message = self.messages_by_transaction[transaction_id]
        message.gateway.direction = "out"
        message.timestamp = time.time()
        new_message = CountMessage(**message.dict())
        self.messages_by_transaction[transaction_id] = new_message
        self.zone_count[new_message.zone] = self.zone_count.get(new_message.zone, 1) - 1
        self.backend.check_out(new_message)
        self.messages_by_transaction.pop(transaction_id)
        return new_message

    def check_out_double_exit(self, **kwargs) -> CountMessage:
        """ check_out call with double exit
        :param kwargs: Parameters for the count message
        """

        kwargs["error"] = "double_exit"
        kwargs["direction"] = "out"
        transaction_id = self.create_count_message(**kwargs)
        message = self.messages_by_transaction[transaction_id]
        message.transaction_id = None
        return self.check_out(transaction_id)

    def check_out_medium_changed(self, transaction_id, medium="nfc", medium_id="mynfcid") -> CountMessage:
        """ check_out call with changed medium
            :param transaction_id: Transaction ID of the check_in message
            :param medium: new medium
            :param medium_id: new medium_id
            """
        message = self.messages_by_transaction[transaction_id]
        assert message.medium_type != medium, "medium needs to change"
        message.error = "medium_type_changed"
        message.previous_id = message.id
        message.previous_medium_type = message.medium_type
        message.medium_type = medium
        message.id = medium_id
        return self.check_out(transaction_id)

    def check_out_medium_id_changed(self, transaction_id, medium_id="mynfcid") -> CountMessage:
        """ check_out call with changed medium id
            :param transaction_id: Transaction ID of the check_in message
            :param medium_id: new medium_id
            """
        message = self.messages_by_transaction[transaction_id]
        assert message.id != medium_id, "medium_id needs to change"
        message.error = "medium_id_changed"
        message.previous_id = message.id
        message.id = medium_id
        return self.check_out(transaction_id)

    def check_out_user_changed(self, transaction_id, user="api-backend_user1") -> CountMessage:
        """ check_out call with changed user
            :param transaction_id: Transaction ID of the check_in message
            :param user: new user
            """
        message = self.messages_by_transaction[transaction_id]
        message.previous_user = message.user
        assert user != message.user, "user needs to change"
        message.user = user
        message.error = "user_changed"
        return self.check_out(transaction_id)

    def check_out_category_changed(self, transaction_id, category="filler-employee") -> CountMessage:
        """ check_out call with changed category
            :param transaction_id: Transaction ID of the check_in message
            :param category: new category
            """
        message = self.messages_by_transaction[transaction_id]
        assert message.category != category, "category needs to change"
        message.category = category
        message.error = "category_changed"
        return self.check_out(transaction_id)



import logging
import os
import socket
import warnings
from types import SimpleNamespace
from typing import Optional, Any
from unittest import TestCase

from pydantic import ValidationError

from openmodule import config
from openmodule.config import validate_config_module
from openmodule.models.base import OpenModuleModel
from openmodule_test.core import OpenModuleCoreTestMixin
from openmodule_test.files import temp_file


class EnvTestCase(TestCase):
    initial_env: dict

    def env(self, value, key="t"):
        os.environ[key] = value

    def setUp(self) -> None:
        self.initial_env = os.environ.copy()
        super().setUp()

    def tearDown(self):
        super().tearDown()
        os.environ.clear()
        os.environ.update(self.initial_env)


class ConfigTest(EnvTestCase):
    def test_int(self):
        self.env("1")
        self.assertEqual(1, config.int("t", 0))

        self.env("a")
        self.assertEqual(0, config.int("t", 0))

        self.env("0x16")
        self.assertEqual(0, config.int("t", 0))

        with self.assertRaises(AssertionError):
            config.int("t", 1.1)

    def test_float(self):
        self.env("1")
        self.assertEqual(1, config.float("t", 0))
        self.env("1.0")
        self.assertEqual(1, config.float("t", 0))
        self.env("1.1")
        self.assertEqual(1.1, config.float("t", 0))

        self.env("a")
        self.assertEqual(0, config.float("t", 0))

        self.env("0x16")
        self.assertEqual(0, config.float("t", 0))

        with self.assertRaises(AssertionError):
            config.float("t", "a")

    def test_str(self):
        self.env("1")
        self.assertEqual("1", config.string("t", "a"))

        self.env(" \"'1' \r\n\t")  # config.string strips unnecessary whitespace and quotes
        self.assertEqual("1", config.string("t", "a"))

        with self.assertRaises(AssertionError):
            config.string("t", 1.0)  # noqa

    def test_bool(self):
        self.env("1")
        self.assertTrue(config.bool("t", False))
        self.env("y")
        self.assertTrue(config.bool("t", False))
        self.env("Y")
        self.assertTrue(config.bool("t", False))
        self.env("t")
        self.assertTrue(config.bool("t", False))
        self.env("true")
        self.assertTrue(config.bool("t", False))
        self.env("yes")
        self.assertTrue(config.bool("t", False))

        # empty string is default value
        self.env("")
        self.assertTrue(config.bool("t", True))
        self.assertFalse(config.bool("t", False))

        with self.assertRaises(AssertionError):
            config.bool("t", "a")  # noqa

    def test_log_level(self):
        self.env("", "LOG_LEVEL")
        self.assertEqual(logging.INFO, config.log_level())  # info is the default value
        self.env("info", "LOG_LEVEL")
        self.assertEqual(logging.INFO, config.log_level())
        self.env("WARN", "LOG_LEVEL")
        self.assertEqual(logging.WARNING, config.log_level())
        self.env("ERROR", "LOG_LEVEL")
        self.assertEqual(logging.ERROR, config.log_level())
        self.env("CRITICAL", "LOG_LEVEL")
        self.assertEqual(logging.CRITICAL, config.log_level())
        self.env("VERBOSE", "LOG_LEVEL")
        self.assertEqual(logging.DEBUG, config.log_level())
        self.env("DEbUG", "LOG_LEVEL")
        self.assertEqual(logging.DEBUG, config.log_level())

    def test_version(self):
        # read version from env first
        self.env("1.0.0", "VERSION")
        self.assertEqual("1.0.0", config.version())

        # if it is not present, read it from VERSION file
        self.env("", "VERSION")
        try:
            with open("VERSION", "w") as f:
                f.write("1.0.0\n")
            self.assertEqual("1.0.0", config.version())
        finally:
            os.unlink("VERSION")

        # if we find nothing, version === "unknown"
        self.assertEqual("unknown", config.version())

    def test_resource(self):
        # no auth credentials set -> use hostname which is nice during development
        self.assertEqual(socket.gethostname(), config.resource())

        self.env("TEST123", "AUTH_RESOURCE")
        self.assertEqual("TEST123", config.resource())


class ConfigYamlTest(EnvTestCase):
    class ExampleModel(OpenModuleModel):
        string: str = "default"

    class NonConstructable(OpenModuleModel):
        string: str

    def test_path_from_env(self):
        with temp_file("string: not-default") as path:
            self.env(path, "CONFIG_YAML")
            instance = config.yaml(self.ExampleModel)
            self.assertEqual('not-default', instance.string)

    def test_default_model(self):
        instance = config.yaml(self.ExampleModel)
        self.assertEqual("default", instance.string)

    def test_invalid_default_model(self):
        # we do not protect against any validation errors, but simply log them
        # in any case this is mostly a developer's fault, and hardly ever a user fault
        with self.assertRaises(ValidationError):
            with self.assertLogs() as cm:
                config.yaml(self.NonConstructable)

        self.assertIn("wrong with the configuration", str(cm.output))

    def test_yaml_ignores_custom_constructors(self):
        yaml = """\
        test: string
        some: !<CustomObject>
            a: b
        """

        class TestModel(OpenModuleModel):
            test: str
            some: Optional[Any]

        with temp_file(yaml) as f:
            model = config.yaml(TestModel, f)

        self.assertIsNone(model.some, msg="Not parsed custom elements must be none")
        self.assertEqual(model.test, "string")


class LegacyConfigSupportTest(EnvTestCase):
    def test_broker_pub(self):
        # nothing found -> defaults
        self.assertEqual("tcp://127.0.0.1:10200", config.broker_pub())

        # legacy configs
        self.env("1.2.3.4", "BROKER_HOST")
        self.env("1234", "BROKER_PUB_PORT")
        self.assertEqual("tcp://1.2.3.4:1234", config.broker_pub())

        # new configs overrule
        self.env("inproc://test-pub", "BROKER_PUB")
        self.env("1.2.3.4", "BROKER_HOST")
        self.env("1234", "BROKER_PUB_PORT")
        self.assertEqual("inproc://test-pub", config.broker_pub())

    def test_broker_sub(self):
        # nothing found -> defaults
        self.assertEqual("tcp://127.0.0.1:10100", config.broker_sub())

        # legacy configs
        self.env("1.2.3.4", "BROKER_HOST")
        self.env("1234", "BROKER_SUB_PORT")
        self.assertEqual("tcp://1.2.3.4:1234", config.broker_sub())

        # new configs overrule
        self.env("inproc://test-pub", "BROKER_SUB")
        self.env("1.2.3.4", "BROKER_HOST")
        self.env("1234", "BROKER_SUB_PORT")
        self.assertEqual("inproc://test-pub", config.broker_sub())

    def test_missing_keys_in_config_cause_assertion(self):
        fake_config = SimpleNamespace()
        fake_config.NAME = "test"
        with self.assertRaises(AssertionError):
            validate_config_module(fake_config)

    def test_legacy_broker_configs_cause_warning(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")

            fake_config = SimpleNamespace()
            fake_config.NAME = "test"
            fake_config.RESOURCE = "test"
            fake_config.VERSION = "test"
            fake_config.DEBUG = False
            fake_config.TESTING = True
            fake_config.BROKER_PUB_PORT = "10000"
            validate_config_module(fake_config)

        self.assertEqual(1, len(w))
        self.assertTrue(issubclass(w[0].category, DeprecationWarning))
        self.assertIn("BROKER_PUB", str(w[0].message))


class MinimalConfig:
    TEST = "abc"
    BEST = "best"


class BaseConfigTest(OpenModuleCoreTestMixin):
    config_kwargs = dict(TEST="asdf", RESOURCE="arivo", NAME="abc")
    base_config = MinimalConfig

    def test_config(self):
        # assert fake config is used
        self.assertIsNotNone(self.core.config.BROKER_PUB)

        # assert init_kwargs overwrite base config
        self.assertEqual(self.core.config.TEST, "asdf")

        # assert init_kwargs overwrite fake config
        self.assertEqual(self.core.config.RESOURCE, "arivo")
        self.assertEqual(self.core.config.NAME, "abc")

        # assert minimal config
        self.assertEqual(self.core.config.BEST, "best")



from logging.config import fileConfig

from alembic import context
from sqlalchemy import engine_from_config
from sqlalchemy import pool

from openmodule.database.database import target_metadata

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.

config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.attributes.get('configure_logging', True):
    fileConfig(config.config_file_name)


# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata


# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.

def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """

    def process_revision_directives(context, revision, directives):
        if config.cmd_opts.autogenerate:
            script = directives[0]
            if script.upgrade_ops.is_empty():
                directives[:] = []

    connectable = config.attributes.get('connection', None)
    if connectable is None:
        connectable = engine_from_config(
            config.get_section(config.config_ini_section),
            prefix="sqlalchemy.",
            poolclass=pool.NullPool,
        )

    with connectable.connect() as connection:
        context.configure(
            connection=connection, target_metadata=target_metadata,
            process_revision_directives=process_revision_directives
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    assert False, "Not in use"
else:
    run_migrations_online()



import logging
from unittest import TestCase

import sentry_sdk

from openmodule.core import OpenModuleCore
from openmodule.sentry import init_sentry
from openmodule_test.zeromq import ZMQTestMixin


class SentryTestCase(ZMQTestMixin, TestCase):
    topics = ["sentry"]

    def setUp(self):
        super().setUp()
        logging.basicConfig(level=logging.INFO)
        self.core = OpenModuleCore(self.zmq_context(), self.zmq_config(DEBUG=False, TESTING=False))
        self.core.start()

    def tearDown(self):
        super().tearDown()
        sentry_sdk.init(dsn=None)  # clears the sentry configuration
        self.core.join()

    def test_sentry_init(self):
        sentry_sdk.capture_message("test message")
        with self.assertRaises(TimeoutError):
            self.zmq_client.wait_for_message_on_topic(b"sentry")

        init_sentry(self.core)
        sentry_sdk.capture_message("test message")

        sentry_message = self.zmq_client.wait_for_message_on_topic(b"sentry")
        self.assertIn("event", sentry_message)

    def test_sentry_logging_integration(self):
        init_sentry(self.core)
        logging.error("Some error, please help!")

        sentry_message = self.zmq_client.wait_for_message_on_topic(b"sentry")
        self.assertIn("event", sentry_message)

    def test_sentry_does_not_run_in_debug(self):
        self.core.config.DEBUG = True
        init_sentry(self.core)
        logging.error("Some error, please help!")
        with self.assertRaises(TimeoutError):
            self.zmq_client.wait_for_message_on_topic(b"sentry")

    def test_extras_are_set(self):
        self.core.config.RESOURCE = "some-test-resource123"
        init_sentry(self.core, extras={"some-more-extras" + "123": True})

        logging.error("Some error, please help!", extra={"even-more-extras" + "123": "yes!"})
        sentry_message = self.zmq_client.wait_for_message_on_topic(b"sentry")

        # we dont know exactly where sentry puts its stuff, so simply check string contains
        event = str(sentry_message["event"])
        self.assertIn("some-test-resource123", event)
        self.assertIn("even-more-extras123", event)
        self.assertIn("some-more-extras123", event)

        # just to be sure, that sentry doesnt put the code here in its event, this would also cause all tests to succeed
        self.assertNotIn("wait_for_message_on_topic", event)



import logging

from openmodule.dispatcher import EventListener, MessageDispatcher
from openmodule.models.base import ZMQMessage


class ConnectionStatusMessage(ZMQMessage):
    connected: bool
    type: str = "connection_status"


class ConnectionStatusListener:
    on_connect: EventListener
    on_disconnect: EventListener
    connected: bool = True
    _first_message: bool = True

    def __init__(self, dispatcher: MessageDispatcher):
        self.log = logging.getLogger(self.__class__.__name__)
        self.on_connect = EventListener(log=self.log)
        self.on_disconnect = EventListener(log=self.log)
        dispatcher.register_handler(b"connection_status",
                                    ConnectionStatusMessage,
                                    self.process_connection_status,
                                    filter=dict(type="connection_status"))

    def process_connection_status(self, message: ConnectionStatusMessage):
        new_connected = message.connected
        if new_connected != self.connected or self._first_message:
            self._first_message = False
            self.connected = new_connected
            if new_connected:
                self.on_connect()
            else:
                self.on_disconnect()



from unittest import TestCase, mock

from openmodule.core import init_openmodule, OpenModuleCore, shutdown_openmodule
from openmodule.dispatcher import MessageDispatcher
from openmodule.utils.connection_status import ConnectionStatusListener, ConnectionStatusMessage
from openmodule_test.eventlistener import MockEvent
from openmodule_test.health import HealthTestMixin
from openmodule_test.zeromq import ZMQTestMixin


class ConnectionStatusTestCase(TestCase):
    on_connect: mock.Mock
    on_disconnect: mock.Mock
    connection_status: ConnectionStatusListener

    def setUp(self):
        self.dispatcher = MessageDispatcher()
        self.connection_status = ConnectionStatusListener(self.dispatcher)

        self.on_connect = mock.Mock(name="on_connect")
        self.on_disconnect = mock.Mock(name="on_disconnect")
        self.connection_status.on_connect.append(self.on_connect)
        self.connection_status.on_disconnect.append(self.on_disconnect)

    def dispatch_status(self, connected: bool):
        self.dispatcher.dispatch(b"connection_status", ConnectionStatusMessage(name="dummy", connected=connected))

    def test_initial_status_is_connected(self):
        self.assertTrue(self.connection_status.connected)

    def test_first_connected_is_always_sent(self):
        # the first on_connect is always sent, because we could startup, not knowing that we are actually disconnected
        self.assertTrue(self.connection_status.connected)
        self.dispatch_status(True)
        self.on_connect.assert_called_once()

        # second on_connect is not sent
        self.dispatch_status(True)
        self.on_connect.assert_called_once()

    def test_event_listener_is_called_once(self):
        self.dispatch_status(False)
        self.on_disconnect.assert_called_once()

        # two connect events, on_connect should only be called once
        self.dispatch_status(True)
        self.dispatch_status(True)
        self.on_connect.assert_called_once()

        # two disconnect events, still on_disconnect should be called once more
        self.dispatch_status(False)
        self.dispatch_status(False)
        self.on_connect.assert_called_once()
        self.assertEqual(self.on_disconnect.call_count, 2)

    def test_connected_member_is_changed(self):
        self.assertTrue(self.connection_status.connected)

        self.dispatch_status(False)
        self.assertFalse(self.connection_status.connected)
        self.on_connect.assert_not_called()
        self.on_disconnect.assert_called_once()

        self.dispatch_status(True)
        self.assertTrue(self.connection_status.connected)
        self.on_connect.assert_called_once()
        self.on_disconnect.assert_called_once()


class ConnectionStatusWithCoreTestCase(HealthTestMixin, ZMQTestMixin, TestCase):
    topics = ["healthz"]
    core: OpenModuleCore

    def setUp(self):
        super().setUp()
        self.core = init_openmodule(self.zmq_config(), context=self.zmq_context())
        self.connection_status = ConnectionStatusListener(self.core.messages)
        self.on_connect = MockEvent(name="on_connect")
        self.on_disconnect = MockEvent(name="on_disconnect")
        self.connection_status.on_connect.append(self.on_connect)
        self.connection_status.on_disconnect.append(self.on_disconnect)
        self.wait_for_dispatcher(self.core.messages)

    def tearDown(self):
        super().tearDown()
        shutdown_openmodule()

    def send_status(self, connected: bool):
        self.zmq_client.send(b"connection_status", ConnectionStatusMessage(connected=connected, name="testclient"))

    def test_connected_member_is_changed(self):
        self.assertTrue(self.connection_status.connected)

        # simulate a disconnect event
        self.on_disconnect.assert_not_called()
        self.send_status(False)
        self.on_disconnect.wait_for_call()
        self.on_connect.assert_not_called()
        self.assertFalse(self.connection_status.connected)

        # simulate a connect event
        MockEvent.reset_all_call_counts()
        self.send_status(True)
        self.on_connect.wait_for_call()
        self.assertTrue(self.connection_status.connected)
        self.on_disconnect.assert_not_called()



import logging
from typing import Optional, Dict, Union


from openmodule.dispatcher import EventListener, MessageDispatcher
from openmodule.models.base import Gateway
from openmodule.models.presence import PresenceBaseMessage, PresenceBackwardMessage, PresenceForwardMessage, \
    PresenceChangeMessage, PresenceLeaveMessage, PresenceEnterMessage
from openmodule.models.vehicle import Vehicle


def vehicle_from_presence_message(message: PresenceBaseMessage):
    return Vehicle(
        id=message.vehicle_id,
        lpr=message.medium.lpr,
        qr=message.medium.qr,
        nfc=message.medium.nfc,
        pin=message.medium.pin,
    )


class PresenceListener:
    on_forward: EventListener[Union[Vehicle, Gateway]]
    on_backward: EventListener[Union[Vehicle, Gateway]]
    on_enter: EventListener[Union[Vehicle, Gateway]]
    on_leave: EventListener[Union[Vehicle, Gateway]]
    on_change: EventListener[Union[Vehicle, Gateway]]

    present_vehicles: Dict[str, Vehicle]

    @property
    def present_vehicle(self) -> Vehicle:
        assert self.gate is not None, (
            "`.present_vehicle` may only be used when listening for a specific gate, this presence listener"
            "listens to all gates, please access the present vehicle per gate via `.present_vehicles[gate]`"
        )
        return self.present_vehicles.get(self.gate)

    def __init__(self, dispatcher: MessageDispatcher, gate: Optional[str] = None):
        self.log = logging.getLogger(self.__class__.__name__)
        self.on_forward = EventListener(log=self.log)
        self.on_backward = EventListener(log=self.log)
        self.on_enter = EventListener(log=self.log)
        self.on_change = EventListener(log=self.log)
        self.on_leave = EventListener(log=self.log)
        self.present_vehicles = dict()
        self.gate = gate

        dispatcher.register_handler(b"presence", PresenceBackwardMessage,
                                    self._on_backward, filter={"type": "backward"})
        dispatcher.register_handler(b"presence", PresenceForwardMessage,
                                    self._on_forward, filter={"type": "forward"})
        dispatcher.register_handler(b"presence", PresenceChangeMessage,
                                    self._on_change, filter={"type": "change"})
        dispatcher.register_handler(b"presence", PresenceLeaveMessage,
                                    self._on_leave, filter={"type": "leave"})
        dispatcher.register_handler(b"presence", PresenceEnterMessage,
                                    self._on_enter, filter={"type": "enter"})

    def _gate_matches(self, message: PresenceBaseMessage):
        return (self.gate is None) or (message.gateway.gate == self.gate)

    def _on_backward(self, message: PresenceBackwardMessage):
        if not self._gate_matches(message):
            return
        vehicle = vehicle_from_presence_message(message)
        self.log.debug("presence backward: %s", vehicle)
        self.on_backward(vehicle, message.gateway)

    def _on_forward(self, message: PresenceForwardMessage):
        if not self._gate_matches(message):
            return
        vehicle = vehicle_from_presence_message(message)
        self.log.debug("presence forward: %s", vehicle)
        self.on_forward(vehicle, message.gateway)

    def _on_leave(self, message: PresenceLeaveMessage):
        if not self._gate_matches(message):
            return
        leaving_vehicle = vehicle_from_presence_message(message)
        self.log.debug("presence leave: %s", leaving_vehicle)
        present_vehicle = self.present_vehicles.get(message.gateway.gate)
        if present_vehicle:
            if present_vehicle.id != leaving_vehicle.id:
                self.log.error("A vehicle left with a different vehicle id than the present one. Tracking is "
                               "inconsistent. We are fake-leaving the currently present vehicle, to ensure consistent "
                               "states.", extra={
                    "present_vehicle": str(present_vehicle),
                    "leaving_vehicle": str(leaving_vehicle)
                })
            self.present_vehicles.pop(message.gateway.gate, None)
            self.on_leave(leaving_vehicle, message.gateway)
        else:
            self.log.error("A vehicle left while non was previously present. Tracking is inconsistent. "
                           "The leave will be ignored, to ensure consistent states.",
                           extra={"leaving_vehicle": str(leaving_vehicle)})

    def _on_enter(self, message: PresenceEnterMessage):
        if not self._gate_matches(message):
            return
        new_vehicle = vehicle_from_presence_message(message)
        self.log.debug("presence enter: %s", new_vehicle)
        present_vehicle = self.present_vehicles.get(message.gateway.gate)
        if present_vehicle:
            self.log.error("A new vehicle entered while one was still present. Tracking is inconsistent. "
                           "A leave will be faked, to ensure consistent states.",
                           extra={
                               "present_vehicle": str(present_vehicle),
                               "new_vehicle": str(new_vehicle)
                           })
            self.present_vehicles.pop(message.gateway.gate, None)
            self.on_leave(present_vehicle, message.gateway)
        self.present_vehicles[message.gateway.gate] = new_vehicle
        self.on_enter(new_vehicle, message.gateway)

    def _on_change(self, message: PresenceChangeMessage):
        if not self._gate_matches(message):
            return
        vehicle = vehicle_from_presence_message(message)
        self.log.debug("presence change: %s", vehicle)
        self.on_change(vehicle, message.gateway)



import time
from unittest import TestCase

from openmodule_test.alert import AlertTestMixin


class AlertTestMixinTest(AlertTestMixin, TestCase):
    topics = ["alert"]

    def setUp(self):
        super().setUp()
        self.config = self.zmq_config()

    def example_alert(self, alert_type="test", status="error"):
        return dict(alert_type=alert_type, handle_type="state_change", status=status,
                    timestamp=time.time(), meta={}, name="om_fancy_test_1")

    def test_any_alert(self):
        self.zmq_client.send(b"alert", type="alert", data=self.example_alert())
        self.assertAlert()

    def test_no_alert(self):
        self.assertNoAlert()

    def test_specific_alert(self):
        alert_1 = self.example_alert()
        alert_2 = self.example_alert(status="ok")
        alert_3 = self.example_alert(alert_type="asdf")

        self.zmq_client.send(b"alert", type="alert", **alert_1)
        self.assertAlert(status="error", alert_type="test")

        self.zmq_client.send(b"alert", type="alert", **alert_2)
        self.assertNoAlert(status="error", alert_type="test")

        self.zmq_client.send(b"alert", type="alert", **alert_3)
        self.assertNoAlert(status="error", alert_type="test")



import contextlib
import os
import tempfile


@contextlib.contextmanager
def temp_file(content):
    path = tempfile.mktemp()
    try:
        with open(path, "w") as f:
            f.write(content)
        yield path
    finally:
        os.unlink(path)



from functools import partial
from typing import Any
from unittest import TestCase

from openmodule.dispatcher import MessageDispatcher
from openmodule.models.base import ZMQMessage


class MessageDispatcherBaseTest(TestCase):
    dispatcher: MessageDispatcher
    message: Any

    def dummy_message(self, type="test", **kwargs):
        return {"type": type, "name": "testclient", **kwargs}

    def _set_true_handler(self, message, var="message"):
        setattr(self, var, message)

    def setUp(self):
        super().setUp()
        self.dispatcher = MessageDispatcher()
        self.message = None


class MessageDispatcherBasicsTestCase(MessageDispatcherBaseTest):
    def test_topic_is_str(self):
        self.dispatcher.register_handler("test", ZMQMessage, self._set_true_handler)
        self.assertIsNone(self.message)

        self.dispatcher.dispatch(b"test", self.dummy_message())
        self.assertIsNotNone(self.message)

    def test_no_filter(self):
        self.dispatcher.register_handler(b"test", ZMQMessage, self._set_true_handler)
        self.assertIsNone(self.message)

        self.dispatcher.dispatch(b"test", self.dummy_message())
        self.assertIsNotNone(self.message)

    def test_filter(self):
        self.dispatcher.register_handler(b"test", ZMQMessage, self._set_true_handler, filter={"type": "some-type"})

        # filter does not match
        self.dispatcher.dispatch(b"test", self.dummy_message(type="incorrect"))
        self.assertIsNone(self.message)

        # filter matches
        self.dispatcher.dispatch(b"test", self.dummy_message("some-type"))
        self.assertIsNotNone(self.message)

    def test_multiple_filter(self):
        # multiple handlers with identical filters and topics are allowed
        self.message1 = None
        self.message2 = None

        self.dispatcher.register_handler(b"test", ZMQMessage, partial(self._set_true_handler, var="message1"),
                                         filter={"type": "some-type"})
        self.dispatcher.register_handler(b"test", ZMQMessage, partial(self._set_true_handler, var="message2"),
                                         filter={"type": "some-type"})

        # filter does not match
        self.dispatcher.dispatch(b"test", self.dummy_message(type="incorrect"))
        self.assertIsNone(self.message1)
        self.assertIsNone(self.message2)

        # filter matches
        self.dispatcher.dispatch(b"test", self.dummy_message("some-type"))
        self.assertIsNotNone(self.message1)
        self.assertIsNotNone(self.message2)

    def test_prefix_does_not_match(self):
        # since zmq subscribes on all topics as prefixes, the dispatcher
        # has to full-match the topic
        self.dispatcher.register_handler(b"test", ZMQMessage, self._set_true_handler)
        self.dispatcher.dispatch(b"testp", self.dummy_message())
        self.assertIsNone(self.message)

        # exact match works
        self.dispatcher.dispatch(b"test", self.dummy_message())
        self.assertIsNotNone(self.message)


class MessageDispatcherWithoutExecutorTestCase(MessageDispatcherBaseTest):
    def test_exception_in_handler(self):
        def raises_exception(message):
            raise Exception("something broke!")

        self.dispatcher.register_handler(b"test", ZMQMessage, raises_exception)
        with self.assertLogs() as cm:
            self.dispatcher.dispatch(b"test", self.dummy_message())
        self.assertIn("something broke!", cm.output[0])

    def test_validation_error(self):
        self.dispatcher.register_handler("test", ZMQMessage, lambda x: None)
        with self.assertLogs() as cm:
            self.dispatcher.dispatch(b"test", {})
        self.assertIn("validation error", cm.output[0])



import logging
import time
from unittest import TestCase

import requests
import requests_mock

from openmodule.utils.api import ApiException, AuthMethod
from openmodule_test.api import ApiMocker
from openmodule.utils.api import Api


class Mocker(ApiMocker):
    host = "http://asdf"

    def test_bad(self):
        self.mocker.get(self.server_url("abc/"), status_code=400)

    def bad_json(self):
        self.mocker.get(self.server_url("abc/"), status_code=200)

    def sleep(self):
        self.mocker.get(self.server_url("abc/"), exc=requests.exceptions.ConnectTimeout)


class ApiExceptionTestCase(TestCase):
    def test(self):
        retry_status = [503, 408, 500, 412]
        other = [400, 404, 300, 403]
        for status in retry_status:
            e = ApiException(status_code=status)
            self.assertEqual(True, e.retry)
        for status in other:
            e = ApiException(status_code=status)
            self.assertEqual(False, e.retry)


class ApiTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.api = Api("http://asdf", AuthMethod("none"), False, timeout=1)
        cls.api1 = Api("http://asdf", AuthMethod("basic"), False, timeout=1)
        cls.api2 = Api("http://asdf", AuthMethod("digest"), False, timeout=1)

    @requests_mock.Mocker(real_http=False)
    def test_methods_ok(self, m):
        api = Mocker(m)
        api.ok()

        self.api.get("abc/")
        self.api.post("abc/")
        self.api.put("abc/")
        self.api.delete("abc/")

    @requests_mock.Mocker(real_http=False)
    def test_methods_bad(self, m):
        api = Mocker(m)
        api.test_bad()
        with self.assertRaises(ApiException) as e:
            self.api.get("abc/")
        self.assertEqual(400, e.exception.status)
        self.assertIn("400", str(e.exception))

        api.timeout()
        with self.assertRaises(ApiException) as e:
            self.api.get("abc/")
        self.assertEqual(408, e.exception.status)
        self.assertIn("408", str(e.exception))

        api.bad_json()
        with self.assertRaises(ApiException) as e:
            self.api.get("abc/")
        self.assertEqual(400, e.exception.status)
        self.assertIn("400", str(e.exception))

        self.api.get("abc/", json_response=False)



"""initial

Revision ID: 3b7d1b53b3c0
Revises: 
Create Date: 2021-02-24 09:13:28.120689

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3b7d1b53b3c0'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('test',
    sa.Column('id', sa.String(), nullable=False),
    sa.Column('value1', sa.Integer(), nullable=True),
    sa.Column('value2', sa.Integer(), nullable=True),
    sa.Column('string', sa.String(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('test')
    # ### end Alembic commands ###



import os
import time
from threading import Thread
from unittest import TestCase

from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound

from openmodule.config import database_folder
from openmodule.database.database import Database, active_databases, database_name
from openmodule.utils.db_helper import update_query, delete_query
from openmodule_test.database import SQLiteTestMixin
from openmodule_test.database_models import DatabaseTestModel


class DatabaseTest(SQLiteTestMixin, TestCase):
    alembic_path = "../tests/test_database_data"

    def test_filter(self):
        data = [DatabaseTestModel(value1=x, value2=x % 3) for x in range(5)]
        with self.database as db:
            db.add_all(data)
            self.assertEqual(1, db.query(DatabaseTestModel).filter_by(value1=0).count())
            self.assertEqual(2, db.query(DatabaseTestModel).filter_by(value2=0).count())

        with self.database as db:
            query = db.query(DatabaseTestModel).filter_by(value2=0)
            res = query.filter_by(value1=0)
            self.assertEqual(1, len(list(res)))
            self.assertEqual(0, res[0].value1)
            self.assertEqual(0, res[0].value2)

    def test_update_filter(self):
        data = [DatabaseTestModel(value1=x, value2=x % 3) for x in range(5)]
        with self.database as db:
            db.add_all(data)

            rows_changed = update_query(db, db.query(DatabaseTestModel).filter_by(value2=2), dict(value1=7))
            self.assertEqual(rows_changed, 1)
            results = db.query(DatabaseTestModel).filter_by(value2=2)
            self.assertEqual(2, results[0].value2)

            rows_changed = update_query(db, db.query(DatabaseTestModel).filter_by(value2=1), dict(value1=7))
            self.assertEqual(rows_changed, 2)
            results = db.query(DatabaseTestModel).filter_by(value2=1)
            self.assertEqual(2, results.count())
            self.assertEqual(3, db.query(DatabaseTestModel).filter_by(value1=7).count())

    def test_delete(self):
        data = [DatabaseTestModel(value1=x % 5, value2=x % 3) for x in range(10)]
        with self.database as db:
            db.add_all(data)

            delete_query(db, db.query(DatabaseTestModel).filter_by(value2=0, value1=17))
            self.assertEqual(10, len(db.query(DatabaseTestModel).all()))

            delete_query(db, db.query(DatabaseTestModel).filter_by(value2=0))
            self.assertEqual(6, len(db.query(DatabaseTestModel).all()))

            delete_query(db, db.query(DatabaseTestModel).filter_by(value2=1, value1=1))
            self.assertEqual(5, len(db.query(DatabaseTestModel).all()))

            tmp = db.query(DatabaseTestModel).filter_by(value1=4)

            self.assertTrue(1, tmp.count())
            db.delete(tmp[0])
            self.assertFalse(any(x.value1 == 4 for x in db.query(DatabaseTestModel).all()))

    def test_rollback(self):
        a = DatabaseTestModel(value1=5, value2=0)
        with self.database as db:
            db.add(a)
            db.flush()
            tmp = a.id
        with self.assertRaises(Exception):
            with self.database as db:
                a.value1 = 6
                db.add(a)
                raise Exception("asdf")

        with self.database as db:
            current = db.query(DatabaseTestModel).get(tmp)
            self.assertEqual(5, current.value1)
            self.assertEqual(1, len(db.query(DatabaseTestModel).all()))

    def test_database_transaction_is_single_threaded(self):
        """
        this test starts a second thread, which waits for 0.5s, after that it tries to aquire the database
        and read the value of an object. The main thread in the meantimes blocks for 2 seconds, and after that
        writes something to the database. Both threads check that the other is at the expected state.
        The test shows that the transaction is single threaded, and all threads are blocked while
        a thread has aquired the db
        """
        main_thread_position = ""
        second_thread_position = ""
        second_thread_value = None

        # initial setup model(1) with value = initial
        with self.database as db:
            model = DatabaseTestModel(id="1", string="initial")
            db.add(model)

        def second_thread():
            nonlocal second_thread_position, second_thread_value
            time.sleep(0.5)
            second_thread_position = "acquiring"
            with self.database as db:
                self.assertEqual(main_thread_position, "end")
                second_thread_model: DatabaseTestModel = db.query(DatabaseTestModel).get("1")
                second_thread_value = second_thread_model.string

        thread = Thread(target=second_thread)
        thread.start()

        with self.database as db:
            time.sleep(2)
            self.assertEqual(second_thread_position, "acquiring")
            model = db.query(DatabaseTestModel).get("1")
            model.string = "changed"
            db.add(model)
            main_thread_position = "end"

        thread.join()
        self.assertEqual(second_thread_value, "changed")

    def test_multiple_databases(self):
        db1 = Database(self.database_folder, name="asdf", alembic_path=self.alembic_path)
        with db1 as db:
            db.add(DatabaseTestModel(value1=0, value2=4))
        with self.assertRaises(Exception) as e:
            db2 = Database(self.database_folder, name="asdf", alembic_path=self.alembic_path)
        self.assertIn("already exists", str(e.exception))
        self.assertTrue(os.path.exists(os.path.join(database_folder(), "asdf.sqlite3")))

    def test_get(self):
        with self.database as db:
            base = db.query(DatabaseTestModel)
            res = base.get("asdf")
            self.assertIsNone(res)

            with self.assertRaises(Exception) as e:
                base.one()
            self.assertEqual(NoResultFound, type(e.exception))

            res = base.one_or_none()
            self.assertIsNone(res)

            res = base.first()
            self.assertIsNone(res)

            model1 = DatabaseTestModel()
            model2 = DatabaseTestModel()
            db.add(model1)
            db.add(model2)

            res = base.first()
            self.assertIsNotNone(res)

            with self.assertRaises(Exception) as e:
                base.one_or_none()
            self.assertEqual(MultipleResultsFound, type(e.exception))

class ShutdownTestCase(TestCase):
    alembic_path = "../tests/test_database_data"

    def get_database(self):
        return Database(database_folder(), name="shutdown", alembic_path=self.alembic_path)

    def check_db_present(self, value):
        tmp = database_name(database_folder(), "shutdown")
        if value:
            self.assertIn(tmp, active_databases.keys())
        else:
            self.assertNotIn(tmp, active_databases.keys())

    def test_shutdown(self):
        self.check_db_present(False)
        database = self.get_database()
        self.check_db_present(True)
        database.shutdown()
        self.check_db_present(False)

        database1 = self.get_database()
        self.check_db_present(True)
        del database1
        self.check_db_present(False)

        def scope():
            database2 = self.get_database()
            self.check_db_present(True)

        scope()
        self.check_db_present(False)

    def test_shutdown_while_using(self):
        database = self.get_database()

        def hold_db():
            with database as db:
                time.sleep(3)

        thread = Thread(target=hold_db)
        thread.start()
        self.assertEqual(True, thread.is_alive())
        self.check_db_present(True)
        database.shutdown()
        self.assertEqual(False, thread.is_alive())
        self.check_db_present(False)

    def test_double_shutdown(self):
        database = self.get_database()
        self.check_db_present(True)

        database.shutdown()

        with self.assertRaises(AssertionError) as e:
            database.shutdown()
        self.assertIn("already closed", str(e.exception))
        self.check_db_present(False)



from urllib.parse import urljoin

import requests
import requests_mock


class ApiMocker:
    host = None

    def __init__(self, mocker):
        self.mocker = mocker
        assert self.host, "you need to set a host"

    def server_url(self, url):
        new_url = urljoin(self.host, url)
        if not new_url.endswith("/") and "?" not in new_url:
            new_url += "/"
        return new_url

    def ok(self):
        self.mocker.get(requests_mock.ANY, status_code=200, json={})
        self.mocker.post(requests_mock.ANY, status_code=200, json={})
        self.mocker.put(requests_mock.ANY, status_code=200, json={})
        self.mocker.delete(requests_mock.ANY, status_code=200, json={})

    def timeout(self):
        self.mocker.get(requests_mock.ANY, exc=requests.exceptions.ConnectTimeout)
        self.mocker.post(requests_mock.ANY, exc=requests.exceptions.ConnectTimeout)
        self.mocker.put(requests_mock.ANY, exc=requests.exceptions.ConnectTimeout)
        self.mocker.delete(requests_mock.ANY, exc=requests.exceptions.ConnectTimeout)

    def unavailable(self):
        self.mocker.get(requests_mock.ANY, status_code=503, json={})
        self.mocker.post(requests_mock.ANY, status_code=503, json={})
        self.mocker.put(requests_mock.ANY, status_code=503, json={})
        self.mocker.delete(requests_mock.ANY, status_code=503, json={})

    def server_error(self):
        self.mocker.get(requests_mock.ANY, status_code=500, json={})
        self.mocker.post(requests_mock.ANY, status_code=500, json={})
        self.mocker.put(requests_mock.ANY, status_code=500, json={})
        self.mocker.delete(requests_mock.ANY, status_code=500, json={})

    def invalid_auth(self):
        self.mocker.get(requests_mock.ANY, status_code=403, json={})
        self.mocker.post(requests_mock.ANY, status_code=403, json={})
        self.mocker.get(requests_mock.ANY, status_code=403, json={})
        self.mocker.post(requests_mock.ANY, status_code=403, json={})

    def bad_request(self):
        self.mocker.get(requests_mock.ANY, status_code=400, json={})
        self.mocker.post(requests_mock.ANY, status_code=400, json={})
        self.mocker.get(requests_mock.ANY, status_code=400, json={})
        self.mocker.post(requests_mock.ANY, status_code=400, json={})

    def not_found(self):
        self.mocker.get(requests_mock.ANY, status_code=404, json={})
        self.mocker.post(requests_mock.ANY, status_code=404, json={})
        self.mocker.get(requests_mock.ANY, status_code=404, json={})
        self.mocker.post(requests_mock.ANY, status_code=404, json={})



from typing import Optional, List

from openmodule.models.base import OpenModuleModel


def _truncate(x):
    if len(x) > 10:
        return x[:10] + "..."
    else:
        return x


class Medium(OpenModuleModel):
    id: str
    type: str

    def __str__(self):
        return f"{self.type}:{_truncate(self.id)}"


class LPRCountry(OpenModuleModel):
    code: str
    state: str = ""
    confidence: float = 1


class LPRAlternative(OpenModuleModel):
    confidence: float
    plate: str


class LPRMedium(Medium):
    type: str = "lpr"
    country: LPRCountry = LPRCountry(code="OTHER")
    confidence: float = 1
    alternatives: List[LPRAlternative] = []

    def __str__(self):
        return f"{self.type}:[{self.country.code}]{_truncate(self.id)}"


class Vehicle(OpenModuleModel):
    id: int
    lpr: Optional[LPRMedium]
    qr: Optional[Medium]
    nfc: Optional[Medium]
    pin: Optional[Medium]

    def __str__(self):
        media = ", ".join(str(x) for x in [self.lpr, self.qr, self.nfc] if x)
        return f"Vehicle<id:{str(self.id)[-7:]}, {media}>"



import logging
from typing import List

from openmodule.core import OpenModuleCore
from openmodule.models.base import ZMQMessage
from openmodule.rpc.server import RPCServer
from openmodule.models.backend import AccessRequest, AccessResponse, CountMessage, MediumAccesses, Access


class Backend:
    """
    Backend template class
    provides basic functionality used for backups
    * subscribes to BackendMessages and automatically registers backend
    * subscribes to CountMessages and calls check_in/check_out correspondingly
    * provides method for the backend / auth rpc with the check_backend_access method

    """

    def __init__(self, core: OpenModuleCore, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.core = core
        self.log = logging.getLogger()

        self.register_at_controller()
        self.core.messages.register_handler(b"count", CountMessage, self.handle_message)
        self.core.messages.register_handler(b"backend", ZMQMessage, self.handle_backend_message)

    def register_rpcs(self, rpc_server: RPCServer):
        rpc_server.add_filter(self._backend_filter, "backend", "auth")
        rpc_server.register_handler("backend", "auth", request_class=AccessRequest,
                                    response_class=AccessResponse, handler=self.rpc_check_access)

    def _backend_filter(self, request, message, handler) -> bool:
        backend = request.name
        if not backend:
            return False
        return self.core.config.NAME == backend

    def check_access(self, request: AccessRequest) -> List[Access]:
        """
        this method checks if current mediums has access to parking lot
        it should raise an Exception if it fails
        :param request: AccessRequest
        :return: Acesses
        """
        raise NotImplementedError()

    def check_in(self, message: CountMessage):
        """
       this method should check in the user of the message in the backend
        it should raise an Exception if it fails
       :param message: CountMessage
       """
        raise NotImplementedError()

    def check_out(self, message: CountMessage):
        """
        this method should check out the user of the message of the backend
        :param message: CountMessage
        """
        raise NotImplementedError()

    def shutdown(self):
        """
        this method should be used if a shutdown routine is required
        """
        pass

    def handle_message(self, message):
        try:
            if message.gateway.direction == "in":
                self.check_in(message)
            else:
                self.check_out(message)
            self.log.info(f"Sending check_{message.gateway.direction} for {message.user} successful")
            return True
        except Exception as e:
            self.log.exception(f"Sending check_{message.gateway.direction} for {message.user} failed")
            return False

    def rpc_check_access(self, request: AccessRequest, _) -> AccessResponse:
        gate_log_string = f"({request.gateway.gate}/{request.gateway.direction})"

        try:
            accesses = self.check_access(request)
        except Exception as e:
            self.log.exception(f"check access request had an internal error {gate_log_string}")
            return AccessResponse(success=False,
                                  medium=MediumAccesses(accesses=[], id=request.id, type=request.medium_type))

        if accesses:
            self.log.info(f"{request.id}:{request.medium_type} has {len(accesses)} permissions {gate_log_string}")
            return AccessResponse(success=True,
                                  medium=MediumAccesses(id=request.id, type=request.medium_type, accesses=accesses))
        else:
            self.log.info(f"{request.id} medium {request.medium_type} has NO parking permissions {gate_log_string}")
            return AccessResponse(success=True,
                                  medium=MediumAccesses(accesses=[], id=request.id, type=request.medium_type))

    def handle_backend_message(self, message: ZMQMessage):
        if message.type.lower() == "register_request":
            self.register_at_controller()

    def register_at_controller(self):
        message = ZMQMessage(name=self.core.config.NAME, type="register")
        self.core.publish(message, b"backend")



import random
from datetime import datetime
from typing import Optional, Callable

from openmodule.models.base import Direction, Gateway
from openmodule.models.presence import PresenceBaseMessage, PresenceEnterMessage, PresenceLeaveMessage, \
    PresenceForwardMessage, PresenceBackwardMessage, PresenceMedia, PresenceChangeMessage
from openmodule.models.vehicle import LPRMedium, LPRCountry, Medium, Vehicle


class VehicleBuilder:
    vehicle_id: int
    medium: PresenceMedia

    def __init__(self):
        self.vehicle_id = random.randint(0, 1000000000000000)
        self.medium = PresenceMedia()

    def vehicle(self) -> Vehicle:
        return Vehicle(
            id=self.vehicle_id,
            lpr=self.medium.lpr,
            qr=self.medium.qr,
            nfc=self.medium.nfc,
            pin=self.medium.pin,
        )

    def id(self, id: int) -> 'VehicleBuilder':
        self.vehicle_id = id
        return self

    def lpr(self, country, plate=None) -> 'VehicleBuilder':
        if country is None:
            self.medium.lpr = None
        else:
            self.medium.lpr = LPRMedium(
                id=plate,
                country=LPRCountry(code=country)
            )
        return self

    def nfc(self, id) -> 'VehicleBuilder':
        if id is None:
            self.medium.nfc = id
        else:
            self.medium.nfc = Medium(id=id, type="nfc")
        return self

    def qr(self, id) -> 'VehicleBuilder':
        if id is None:
            self.medium.qr = id
        else:
            self.medium.qr = Medium(id=id, type="qr")
        return self

    def pin(self, id) -> 'VehicleBuilder':
        if id is None:
            self.medium.pin = id
        else:
            self.medium.pin = Medium(id=id, type="pin")
        return self


class PresenceSimulator:
    current_present: Optional[VehicleBuilder] = None

    def __init__(self, gate: str, direction: Direction, emit: Callable[[PresenceBaseMessage], None]):
        self.gateway = Gateway(gate=gate, direction=direction)
        self.emit = emit

    def vehicle(self):
        return VehicleBuilder()

    def _common_kwargs(self, vehicle):
        timestamp = datetime.now().timestamp()
        return {
            "vehicle_id": vehicle.vehicle_id,
            "unsure": False,
            "present-area-name": f"{self.gateway.gate}-present",
            "last_update": timestamp,
            "name": "presence-sim",
            "source": self.gateway.gate,
            "gateway": self.gateway,
            "medium": vehicle.medium
        }

    def enter(self, vehicle: VehicleBuilder):
        if self.current_present:
            self.leave()
        self.current_present = vehicle
        self.emit(PresenceEnterMessage(**self._common_kwargs(vehicle)))

    def leave(self):
        self.emit(PresenceLeaveMessage(**self._common_kwargs(self.current_present)))
        temp = self.current_present
        self.current_present = None
        return temp

    def forward(self, vehicle: Optional[VehicleBuilder] = None):
        assert vehicle or self.current_present, "a vehicle must be present, or you have to pass a vehicle"
        if not vehicle:
            vehicle = self.leave()
        self.emit(PresenceForwardMessage(
            **self._common_kwargs(vehicle),
            **{"leave-time": datetime.now().timestamp()}
        ))

    def backward(self, vehicle: Optional[VehicleBuilder] = None):
        assert vehicle or self.current_present, "a vehicle must be present, or you have to pass a vehicle"
        if not vehicle:
            vehicle = self.leave()
        self.emit(PresenceBackwardMessage(
            **self._common_kwargs(vehicle),
            **{"leave-time": datetime.now().timestamp()}
        ))

    def change(self, vehicle: Optional[VehicleBuilder] = None):
        assert vehicle or self.current_present, "a vehicle must be present, or you have to pass a vehicle"
        self.emit(PresenceChangeMessage(
            **self._common_kwargs(vehicle),
            **{"leave-time": datetime.now().timestamp()}
        ))






import traceback

import os
import sys
from functools import partial

import logging
import time
from zmq import ContextTerminated


def _thread_wrapper(*args, _run_function, _return_code, **kwargs):
    try:
        _run_function(*args, **kwargs)
    except (KeyboardInterrupt, ContextTerminated):  # pragma: no cover
        pass
    except Exception as e:  # pragma: no cover
        logging.exception(e)
        sys.stderr.flush()
        time.sleep(1)
        os._exit(_return_code)


def get_thread_wrapper(function, return_code=10):
    return partial(_thread_wrapper, _run_function=function, _return_code=return_code)



from unittest import TestCase

from openmodule.core import OpenModuleCore
from openmodule.health import health_ok, health_error
from openmodule_test.health import HealthTestMixin


class HealthTestCase(HealthTestMixin, TestCase):
    topics = ["healthz"]

    def setUp(self):
        super().setUp()
        self.core = OpenModuleCore(self.zmq_context(), self.zmq_config())
        self.core.start()
        self.wait_for_health()

    def tearDown(self):
        super().tearDown()
        self.core.join()

    def test_health_ok(self):
        self.core.health.health_handler = lambda: health_ok({"some": "meta"})
        health = self.get_health()
        self.assertHealthOk(health)

    def test_health_error(self):
        self.core.health.health_handler = lambda: health_error("some error message")
        health = self.get_health()
        self.assertHealthError(health)

    def test_health_handler_error(self):
        self.core.health.health_handler = lambda: 1 / 0
        health = self.get_health()
        self.assertHealthError(health)
        self.assertIn("health routine", str(health))

    def test_health_handler_invalid_return_type(self):
        self.core.health.health_handler = lambda: None
        health = self.get_health()
        self.assertHealthError(health)
        self.assertIn("health routine", str(health))



from typing import Optional

from openmodule_test.zeromq import ZMQTestMixin


def _get_health_filter(name):
    return lambda topic, message: (topic == b"healthz") and \
                                  ("pong" in message) and \
                                  (name is None or message.get("name") == name)


class HealthTestMixin(ZMQTestMixin):
    def get_health(self, name: Optional[str] = None, timeout: float = 3):
        self.assertSubscription(b"healthz")
        self.zmq_client.send("healthz", {"type": "ping", "ping": "ping"})
        _, message = self.zmq_client.wait_for_message(
            _get_health_filter(name),
            timeout=timeout
        )
        return message

    def assertHealthOk(self, message):
        self.assertIn("pong", message)
        pong = message.get("pong", {})
        self.assertEqual("ok", pong.get("status"))

    def assertHealthError(self, message):
        self.assertIn("pong", message)
        pong = message.get("pong", {})
        self.assertEqual("error", pong.get("status"))

    def wait_for_health(self, name: Optional[str] = None):
        """
        :param name: if specified the function waits for a specific service name on startup
        """

        for x in range(self.zmq_client.startup_check_iterations):
            try:
                self.get_health(name, timeout=self.zmq_client.startup_check_delay)
            except TimeoutError:
                pass
            else:
                return

        assert False, f"health did not answer within " \
                      f"{self.zmq_client.startup_check_delay * self.zmq_client.startup_check_iterations} seconds"



from typing import Optional, Dict, Any
from uuid import UUID

from openmodule.models.base import ZMQMessage


class RPCRequest(ZMQMessage):
    rpc_id: UUID
    request: Optional[Dict]


class RPCResponse(ZMQMessage):
    rpc_id: Optional[UUID]
    response: Any



from datetime import datetime, timedelta
from typing import Optional
from unittest import TestCase

import orjson
from pydantic import ValidationError

from openmodule.models.base import OpenModuleModel, ZMQMessage, Gateway


class TestModel(OpenModuleModel):
    __test__ = False
    value: str


class ModelTestCase(TestCase):
    def test_json_is_forbidden(self):
        instance = TestModel(value="Hello World")
        with self.assertRaises(AssertionError):
            instance.json()

    def test_json_dumps_and_load(self):
        instance = TestModel(value="Hello World")
        instance2 = TestModel.parse_raw(instance.json_bytes())
        self.assertEqual(instance.value, instance2.value)

    def test_unset_default_are_serialized(self):
        class SomeZMQMessage(ZMQMessage):
            type: str = "my-type"

        msg = SomeZMQMessage(name="myname").json_bytes()
        data = orjson.loads(msg)

        self.assertIn("type", data)
        self.assertEqual("my-type", data.get("type"))

    def test_none_values_are_not_exported(self):
        class SomeZMQMessage(ZMQMessage):
            type: str = "test"
            value: Optional[float]

        data = orjson.loads(SomeZMQMessage(name="myname").json_bytes())
        self.assertNotIn("value", data)

        data = orjson.loads(SomeZMQMessage(name="myname", value=1).json_bytes())
        self.assertIn("value", data)
        self.assertEqual(1, data["value"])

        # to my understanding pydantic, openapi and jsonschema all do not agree how nullable but required
        # fields should correctly be handled / annotated. Currently our strategy is to use exclude_none=True
        # so null values are not exported, in an ideal world i would like to have value=None exported, but
        # it is currently not possible
        data = orjson.loads(SomeZMQMessage(name="myname", value=None).json_bytes())
        # self.assertIn("value", data)
        # self.assertEqual(None, data["value"])
        self.assertNotIn("value", data)


class ZMQMessageTestCase(TestCase):
    def test_timestamp_to_datetime(self):
        instance = ZMQMessage(timestamp=1607586354631, name="test", type="test")
        self.assertEqual(instance.timestamp.isoformat(), "2020-12-10T07:45:54.631000+00:00")

    def test_timestamp_is_now_if_unset(self):
        instance = ZMQMessage(name="test", type="test")
        self.assertAlmostEqual(instance.timestamp, datetime.now(), delta=timedelta(seconds=1))

    def test_datetime_is_serialized_to_unix_timestamp(self):
        instance = ZMQMessage(name="test", type="test")
        json = instance.json_bytes()

        loaded_json = orjson.loads(json)
        self.assertIsInstance(loaded_json["timestamp"], float)
        self.assertAlmostEqual(loaded_json["timestamp"], datetime.now().timestamp(), delta=1)

    def test_receive_optional_gateway(self):
        class OptionalGateway(ZMQMessage):
            type = "something"
            gateway: Gateway = Gateway()

        message = OptionalGateway(name="test")
        self.assertEqual(message.gateway.direction, "")
        self.assertEqual(message.gateway.gate, "")

        message = OptionalGateway(name="test", gateway={"direction": "in"})
        self.assertEqual(message.gateway.direction, "in")
        self.assertEqual(message.gateway.gate, "")

        with self.assertRaises(ValidationError)  as e:
            OptionalGateway(name="test", gateway={"direction": "incorrect"})

        errors = e.exception.errors()
        self.assertEqual(errors[0]["loc"], ("gateway", "direction"))



from __future__ import unicode_literals

import logging
import random
import socket
import string
import threading
import types
from _weakrefset import WeakSet
from contextlib import closing
from enum import Enum
from typing import List, Tuple, Optional, Union, Dict
from unittest import TestCase
from uuid import uuid4, UUID

import orjson
import time
import zmq
from pydantic import BaseModel, Field


def patch_bind_string(bind_string: str) -> str:
    """
    patches bind strings from zmq.Socket.bind so they can be used with zmq.Socket.connect
    """
    return bind_string.replace("://*", "://127.0.0.1")


class TestBroker(threading.Thread):
    __test__ = False

    _termination_watcher_timeout = 10

    xsub: zmq.Socket = None
    xpub: zmq.Socket = None

    def __init__(self, sub_bind, pub_bind):
        super().__init__()
        self.log = logging.getLogger(self.__class__.__name__)
        self.sub_bind = sub_bind
        self.pub_bind = pub_bind
        self.running = False
        self.context = zmq.Context()

    def start(self) -> None:
        # overwriting start ensures that the sockets are bound after calling start()
        # if we bind in run(...) we might bind after another thread, and in the constructor
        # may be surprising to someone using the class aswell
        self.running = True
        self.xsub = self.context.socket(zmq.XSUB)
        self.xpub = self.context.socket(zmq.XPUB)
        self.xsub.setsockopt(zmq.LINGER, 0)
        self.xpub.setsockopt(zmq.LINGER, 0)
        self.log.debug(f"binding on sub->{self.sub_bind} and pub->{self.pub_bind}")
        self.xsub.bind(self.sub_bind)
        self.xpub.bind(self.pub_bind)
        super().start()

    def run(self):
        try:
            self.log.debug("proxy started")
            zmq.proxy(self.xsub, self.xpub)
        except zmq.ContextTerminated:
            self.log.debug("broker thread received context terminated")
        self.xsub.close()
        self.xpub.close()

    def _detect_and_close_bad_linger(self):
        sockets = self.context._sockets  # noqa, modified version of context.destroy() has to use private member
        self.context._sockets = WeakSet()
        has_bad_linger = False
        for s in sockets:
            closed = False
            if s and not s.closed:
                linger = s.getsockopt(zmq.LINGER)
                if linger is None or linger == -1:
                    has_bad_linger = True
                    s.close(linger=0)
                    closed = True

            if not closed:  # we may need them later, for forcefully closing in the termination watcher
                self.context._sockets.add(s)
        return has_bad_linger

    def stop(self):
        terminated = False
        had_open_sockets = False

        def termination_watcher():
            nonlocal had_open_sockets
            for x in range(self._termination_watcher_timeout * 10):
                if terminated:
                    return
                else:
                    time.sleep(0.1)

            if not terminated:
                had_open_sockets = True
                sockets = self.context._sockets  # noqa
                self.context._sockets = WeakSet()
                for socket in sockets:
                    if socket and not socket.closed:
                        socket.close()

        has_bad_linger = self._detect_and_close_bad_linger()
        watcher = threading.Thread(target=termination_watcher)
        watcher.start()
        self.context.term()
        terminated = True
        self.join()

        assert not has_bad_linger, (
            "you have a socket with infinite linger time, this can cause blocking during shutdown, preventing\n"
            "your application from ever shutting down. Please specify a linger timeout via setsockopt(...)\n"
            "or use get_(s/p)ub_socket(...)"
        )

        assert not had_open_sockets, (
            "You have open sockets which did not close after 10 seconds. Some component in your testcase\n"
            "did not shutdown properly. I am forcefully closing the sockets from a different thread to\n"
            "prevent job from getting stuck for hours in the CI/CD. This can cause various internal ZMQ errors."
        )

        self.log.debug("proxy stopped")


class TestClient(threading.Thread):
    __test__ = False

    startup_check_delay = 0.05
    startup_check_iterations = 20  # 40 * 0.05 = 2 seconds

    _command_topic: bytes
    subscribed_topics: set
    pending_topics: Dict[bytes, None]

    pub: zmq.Socket = None
    sub: zmq.Socket = None

    def __init__(self, broker: TestBroker):
        super().__init__()
        self._command_topic = b"_testclient_" + str(uuid4()).encode("ascii")
        self.subscribed_topics = set()
        self.pending_topics = dict()

        self.broker = broker
        self.connected = threading.Event()
        self.log = logging.getLogger(self.__class__.__name__)
        self.recv_lock = threading.Lock()
        self.recv_messages = []
        self.has_messages = threading.Event()
        self.receiving_thread_id = None
        self.send_lock = threading.Lock()
        self.running = True

    def subscribe(self, *topics: bytes):
        for topic in topics:
            self.sub.subscribe(topic)
            self.subscribed_topics.add(topic)
            self.pending_topics[topic] = None
        self.wait_for_topics()

    def wait_for_topics(self):
        for x in range(self.startup_check_iterations):
            with self.recv_lock:
                pending_topics = list(self.pending_topics.keys())
            if pending_topics:
                for topic in pending_topics:
                    self._zmq_cmd("hi", topic=topic)
                time.sleep(self.startup_check_delay)
            else:
                break

        assert not self.pending_topics, "error during startup and connect"

    def start(self):
        self.pub = self.broker.context.socket(zmq.PUB)
        self.sub = self.broker.context.socket(zmq.SUB)
        self.pub.setsockopt(zmq.LINGER, 0)
        self.sub.setsockopt(zmq.LINGER, 0)
        self.pub.connect(patch_bind_string(self.broker.sub_bind))
        self.sub.connect(patch_bind_string(self.broker.pub_bind))
        super().start()
        self.subscribe(self._command_topic)

    def _zmq_cmd(self, cmd, topic=None):
        if not self.broker.context.closed:
            message = {"__testcommand": cmd, "name": "testclient", "type": "cmd"}
            self.pub.send_multipart((topic or self._command_topic, orjson.dumps(message)))

    def run(self):
        try:
            while self.running:
                try:
                    topic, message = self.sub.recv_multipart()
                    message = orjson.loads(message)
                except (KeyError, TypeError, ValueError) as e:
                    self.log.exception("Received an invalid message on the message queue")
                    raise e from None

                if topic == self._command_topic or "__testcommand" in message:
                    if message["__testcommand"] == "exit":
                        break
                    else:
                        with self.recv_lock:
                            if topic in self.pending_topics:
                                del self.pending_topics[topic]
                        continue

                with self.recv_lock:
                    self.recv_messages.append((topic, message))
                    self.has_messages.set()
        except zmq.ContextTerminated:
            pass
        except Exception:  # pragma: no cover
            self.log.error("Internal exception, shutting down")
        else:
            self.log.debug("client thread stopped gracefully")
        finally:
            self.sub.close()
            self.pub.close()

    def send(self, topic: Union[bytes, str], _message=None, **kwargs):
        """
        sends a message on the topic
        :param topic:
        :param _message:
        :param kwargs:
        :return:
        """
        if hasattr(topic, "encode"):
            topic = topic.encode("ascii")

        if isinstance(_message, BaseModel):
            with self.send_lock:
                self.pub.send_multipart((topic, orjson.dumps(_message.dict())))
        else:
            assert not (_message and kwargs), (
                "pass the message dict as the first parameter, or use the kwargs, not both"
            )
            data = _message or kwargs
            assert "type" in data, "a message must always have a type"
            data.setdefault("name", "testcase")
            data.setdefault("timestamp", time.time())
            with self.send_lock:
                self.pub.send_multipart((topic, orjson.dumps(data)))

    def _zmq_pop_from_front(self):
        with self.recv_lock:
            msg = self.recv_messages.pop(0)
            if not self.recv_messages:
                self.has_messages.clear()
            return msg

    def wait_for_message(self, filter, timeout=3) -> Tuple[bytes, dict]:
        """
        :param filter: filter function of type (topic: bytes, message: dict) -> bool
        :return: tuple containing [topic, message]
        """

        # protect the developer from using the client in multiple threads, this is not supported
        with self.recv_lock:
            assert self.receiving_thread_id is None, (
                "the test client is not thread safe! you have to use a separate client for each thread which"
                "wants to receive messages"
            )
            self.receiving_thread_id = threading.get_ident()

        try:
            start = time.time()
            while True:
                if not self.has_messages.wait(timeout=timeout):
                    raise TimeoutError()

                while self.recv_messages:
                    message_topic, message = self._zmq_pop_from_front()
                    if filter(message_topic, message):
                        return message_topic, message

                time_diff = time.time() - start
                if time_diff > timeout:
                    raise TimeoutError()
        finally:
            self.receiving_thread_id = None

    def wait_for_message_on_topic(self, topic: Union[str, bytes], timeout=3) -> dict:
        """
        :return: tuple containing [topic, message]
        """
        if topic and hasattr(topic, "encode"):
            topic = topic.encode()

        assert topic in self.subscribed_topics, (
            "please subscribe to the topic you want to receive from first! call:\n"
            f'  > zmq_client.subscribe("{topic.decode()}")\n'
            f'or set\n'
            f'  > topics = ["{topic.decode()}"] in your test class'
        )
        return self.wait_for_message(lambda recv_topic, _: (not topic) or recv_topic == topic, timeout=timeout)[1]

    def stop(self):
        self._zmq_cmd("exit")
        self.join()


def fake_config(broker: Optional[TestBroker] = None, **kwargs):
    result = {
        "NAME": "test",
        "RESOURCE": "test-resource",
        "VERSION": "test-version",
        "LOG_LEVEL": logging.DEBUG,
        "TESTING": True,
        "DEBUG": False
    }
    if broker:
        result["BROKER_SUB"] = patch_bind_string(broker.sub_bind)
        result["BROKER_PUB"] = patch_bind_string(broker.pub_bind)

    result.update(kwargs)

    # converts to an object
    config = types.SimpleNamespace()
    for k, v in result.items():
        setattr(config, k, v)
    return config


class _TestRPCRequest(BaseModel):
    """
    we do not want to depend on openmodule, this is a minimal version of the zmq message for the rpc function
    """
    timestamp: float = Field(default_factory=time.time)
    name: str
    type: str
    rpc_id: UUID
    request: Optional[Dict]


class ZMQProcotol(str, Enum):
    inproc = "inproc://"
    tcp = "tcp://"


def _find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('127.0.0.1', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


class ZMQTestMixin(TestCase):
    topics: List[Union[bytes, str]] = []
    rpc_channels: List[Union[bytes, str]] = []
    protocol: ZMQProcotol = "inproc://"

    zmq_broker: TestBroker
    zmq_client: TestClient

    def setUp(self):
        super(ZMQTestMixin, self).setUp()

        assert self.protocol in ["inproc://", "tcp://"]
        if self.protocol == "inproc://":
            random_suffix = "".join(random.choices(string.ascii_letters, k=10))
            sub = f"inproc://test-sub-{random_suffix}"
            pub = f"inproc://test-pub-{random_suffix}"
        else:
            sub = f"tcp://127.0.0.1:{_find_free_port()}"
            pub = f"tcp://127.0.0.1:{_find_free_port()}"

        self.zmq_broker = TestBroker(sub, pub)
        self.zmq_broker.start()
        self.zmq_client = TestClient(self.zmq_broker)
        self.zmq_client.start()

        topics = set(self.topics + [f"rpc-rep-{x}" for x in self.rpc_channels])
        topics = [x.encode("ascii") if isinstance(x, str) else x for x in topics]
        self.zmq_client.subscribe(*topics)

    def wait_for_dispatcher(self, dispatcher):
        received = False

        def handler(_):
            nonlocal received
            received = True

        """
        waits until a message dispatcher receives messages, this assumes that the subscription we issue is the last
        and if it is connected, all previous subscriptions will also be connected
        """
        random_topic = "_test" + "".join(random.choices(string.ascii_letters, k=10))
        random_topic = random_topic.encode()
        dispatcher.register_handler(random_topic, BaseModel, handler)

        for x in range(self.zmq_client.startup_check_iterations):
            self.zmq_client.send(random_topic, {"type": "connection-check"})
            time.sleep(self.zmq_client.startup_check_delay)
            if received:
                break
        assert received, "error during startup and connect"

    def zmq_context(self):
        return self.zmq_broker.context

    def zmq_config(self, **kwargs):
        return fake_config(self.zmq_broker, **kwargs)

    def tearDown(self):
        assert hasattr(self, "zmq_client"), (
            "testcase has no zmq_client member, did you forget to call super().setUp()?"
        )
        super(ZMQTestMixin, self).tearDown()
        self.zmq_client.stop()
        self.zmq_broker.stop()

    def rpc(self, channel: str, type: str, request, timeout=3, receive_response=True) -> Optional[dict]:
        response_topic = f"rpc-rep-{channel}".encode("ascii")
        assert response_topic in self.zmq_client.subscribed_topics, (
            "you have to list the rpc channels you want to use beforehand. please set:\n"
            f'  > rpc_channels = ["{channel}"] in your test class '
        )

        rpc_id = str(uuid4())
        rpc_request = _TestRPCRequest(name="testclient", type=type, rpc_id=rpc_id, request=request)
        self.zmq_client.send(f"rpc-req-{channel}".encode("ascii"), rpc_request)
        if receive_response:
            _, response = self.zmq_client.wait_for_message(
                filter=lambda topic, message: topic == response_topic and message.get("rpc_id") == rpc_id,
                timeout=timeout
            )
            return response
        else:
            return None

    def assertRPCFailure(self, response: dict, expected_status="error"):
        self.assertEqual(expected_status, response.get("response", {}).get("status"),
                         msg=f"Expected status to not be ok, got {response.get('response')}")

    def assertRPCSuccess(self, response: dict):
        self.assertEqual("ok", response.get("response", {}).get("status"),
                         msg=f"Expected status to be ok, got {response.get('response')}")

    def assertSubscription(self, *topics: bytes):
        assert all((x in self.zmq_client.subscribed_topics) for x in topics), (
            f"some test functions require topics {b', '.join(topics).decode()}, please set "
            f"""topics = [{', '.join(f'"{x}"' for x in topics)}] in your test class"""
        )



from unittest import TestCase

from openmodule.alert import AlertHandleType
from openmodule.core import core, init_openmodule, shutdown_openmodule
from openmodule_test.alert import AlertTestMixin
from openmodule_test.core import OpenModuleCoreTestMixin
from openmodule_test.health import HealthTestMixin


class OpenModuleCoreTest(AlertTestMixin, HealthTestMixin, TestCase):
    topics = ["healthz", "sentry", "alert"]
    protocol = "tcp://"

    def test_core_init_and_shutdown(self):
        self.assertIsNone(core())

        init_openmodule(self.zmq_config(NAME="x123x"))
        try:
            self.assertIsNotNone(core())
            self.wait_for_health(name="x123x")
        finally:
            shutdown_openmodule()

    def test_core_sentry_error_logging(self):
        # sentry is not active during unittests / debug
        init_openmodule(self.zmq_config(DEBUG=False, TESTING=False))
        self.wait_for_health()

        try:
            core().log.error("some-error-message")
            sentry_message = self.zmq_client.wait_for_message_on_topic("sentry")
            self.assertIn("some-error-message", str(sentry_message))
        finally:
            shutdown_openmodule()

    def test_core_alerts(self):
        init_openmodule(self.zmq_config())
        self.wait_for_health()

        try:
            core().alerts.send("some-source", "some_type", AlertHandleType.state_change)
            self.assertAlert(alert_type="some_type")
        finally:
            shutdown_openmodule()



from unittest import TestCase

from openmodule.alert import AlertHandler, AlertHandleType
from openmodule.core import OpenModuleCore
from openmodule_test.alert import AlertTestMixin


class AlertTestCase(AlertTestMixin, TestCase):
    topics = ["alert"]

    def setUp(self):
        super().setUp()
        self.core = OpenModuleCore(self.zmq_context(), self.zmq_config())
        self.core.start()

    def tearDown(self):
        super().tearDown()
        self.core.join()

    def test_value_is_required_for_state(self):
        alert_handler = AlertHandler(self.core)
        alert_handler.send("pin1", "some_alert", AlertHandleType.state, value=10)
        self.assertAlert()

        with self.assertRaises(ValueError) as e:
            alert_handler.send("pin1", "some_alert", AlertHandleType.state)
        self.assertIn("value must not be none", str(e.exception).lower())

        self.assertNoAlert()

    def test_value_none_is_not_sent(self):
        alert_handler = AlertHandler(self.core)
        alert_handler.send("pin1", "some_alert", AlertHandleType.state_change)
        alert = self.assertAlert(alert_type="some_alert")

        self.assertNotIn("value", alert)



import logging
import os
import shutil
import threading
from typing import Optional

import zmq

from openmodule.alert import AlertHandler
from openmodule.config import validate_config_module
from openmodule.database.database import Database
from openmodule.dispatcher import SubscribingMessageDispatcher
from openmodule.health import HealthHandlerType, Healthz, HealthPingMessage
from openmodule.logging import init_logging
from openmodule.messaging import get_pub_socket, get_sub_socket, receive_message_from_socket
from openmodule.models.base import ZMQMessage
from openmodule.sentry import init_sentry
from openmodule.threading import get_thread_wrapper


class OpenModuleCore(threading.Thread):
    database = None

    def __init__(self, context, config):
        super().__init__(target=get_thread_wrapper(self._run))
        self.context = context
        self.config = config

        self.pub_lock = threading.Lock()
        self.sub_lock = threading.Lock()

        self.pub_socket = get_pub_socket(self.context, self.config, linger=1000)
        self.sub_socket = get_sub_socket(self.context, self.config)
        self.messages = SubscribingMessageDispatcher(self._subscribe)

        self.log = logging.getLogger(self.__class__.__name__)
        self.health = Healthz(self)
        self.alerts = AlertHandler(self)
        self.messages.register_handler(b"healthz", HealthPingMessage, self.health.process_message,
                                       filter={"type": "ping"})

    def _subscribe(self, topic: bytes):
        with self.sub_lock:
            self.sub_socket.subscribe(topic)

    def init_database(self):
        if self.config.TESTING:
            self.database = Database(self.config.DATABASE_FOLDER, self.config.NAME, alembic_path="../src/database")
        else:
            self.database = Database(self.config.DATABASE_FOLDER, self.config.NAME)

    def _run(self):
        try:
            while True:
                topic, message = receive_message_from_socket(self.sub_socket)
                self.messages.dispatch(topic, message)
        except zmq.ContextTerminated:
            self.log.debug("context terminated, core shutting down")
        finally:
            self.pub_socket.close()
            self.sub_socket.close()

    def publish(self, message: ZMQMessage, topic: bytes):
        with self.pub_lock:
            message.publish_on_topic(self.pub_socket, topic)


_core_thread: Optional[OpenModuleCore] = None


def init_dsgvo_config():
    if os.path.exists("/data") and os.path.exists("/data/config.yml"):
        shutil.copyfile("/data/config.yml", "/data/tmp.yml")
        shutil.move("/data/tmp.yml", "/data/dsgvo-config.yml")
    if os.path.exists("/data") and os.path.exists("dsgvo-default.yml"):
        shutil.copyfile("dsgvo-default.yml", "/data/tmp.yml")
        shutil.move("/data/tmp.yml", "/data/dsgvo-default.yml")


def print_environment(core: OpenModuleCore):
    core.log.info(f"Service: {core.config.NAME} (version:{core.config.VERSION})")
    if core.config.DEBUG:
        core.log.warning(
            "\n"
            "       DEBUG MODE is active.\n"
            "       Deactivate by setting environment variable DEBUG=False.\n"
            "       Debug is disabled per default when a version string is set or ran inside docker.\n"
        )


def init_openmodule(config, sentry=True, logging=True, dsgvo=True,
                    health_handler: Optional[HealthHandlerType] = None,
                    context=None, database=False) -> OpenModuleCore:
    context = context or zmq.Context()
    validate_config_module(config)

    global _core_thread
    assert not _core_thread, "openmodule core already running"
    _core_thread = OpenModuleCore(context, config)
    _core_thread.start()

    if logging:
        init_logging(_core_thread)
    print_environment(_core_thread)

    if sentry:
        init_sentry(_core_thread)

    if dsgvo:
        init_dsgvo_config()

    if health_handler:
        _core_thread.health.health_handler = health_handler

    if database:
        _core_thread.init_database()

    return _core_thread


def core() -> OpenModuleCore:
    return _core_thread


def shutdown_openmodule():
    global _core_thread
    assert _core_thread, "core thread is not running, did you call init_openmodule(...)?"

    current_cure = _core_thread
    _core_thread = None

    shutdown_done = threading.Event()

    def last_will():
        if not shutdown_done.wait(timeout=1):
            os._exit(99)

    if not (current_cure.config.TESTING or current_cure.config.DEBUG):
        last_will_thread = threading.Thread(target=last_will)
        last_will_thread.setDaemon(True)
        last_will_thread.start()

    if current_cure.database:
        current_cure.database.shutdown()

    current_cure.context.term()
    current_cure.join()
    shutdown_done.set()



import time
from _weakrefset import WeakSet
from unittest.mock import Mock

_all_mocks = WeakSet()


class MockEvent(Mock):
    _wait_sleep = 0.05
    _wait_called_since_reset = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _all_mocks.add(self)

    @classmethod
    def reset_all_call_counts(cls):
        for x in _all_mocks:
            x.reset_call_count()

    def reset_call_count(self):
        self.call_count = 0
        self._wait_called_since_reset = False

    def wait_for_call(self, timeout=3, minimum_call_count=1):
        assert not self._wait_called_since_reset, (
            "The test-case MUST reset the call count to zero on its own before calling wait_for_call.\n"
            "This is in order to prevent a race-condition which is impossible to prevent otherwise.\n"
            "Please call `.reset_call_count()` or `MockEvent.reset_all_call_counts()`"
        )
        """
        Explanation: Take the following code for example
        ```
        1:  self.zmq_client.send("some-message")
        2:  self.my_event.wait_for_call()
        3:  self.zmq_client.send("some-message")
        4:  self.my_event.wait_for_call()
        ```
        If the unit-test thread is scheduled away after 3, and the mock event is called before 4
        the mock event CANNOT possibly know if it's call count is at 1 because it has been called
        by the event triggered in line 1, or by the event triggered in line 3. Also resetting the 
        call count at the start of `.wait_for_call()` would not help, because in line 4, the call
        count is already at 2, and thus producing an incorrect timeout error.
        
        Only the developer knows what the earliest possible time is that ane vent can trigger, and
        then has to reset the call count one line before. In the case above between line 2 and 3
        """
        self._wait_called_since_reset = True
        for x in range(int(timeout // self._wait_sleep)):
            time.sleep(self._wait_sleep)
            if self.call_count >= minimum_call_count:
                return
        self.assert_called()



from unittest import TestCase

from openmodule.utils.charset import Charset, CharsetConverter, legacy_lpr_charset


class CharsetModelTest(TestCase):
    def test_uppercase_on_construct(self):
        charset = Charset(allowed="abc")
        self.assertEqual("ABC", charset.allowed)

    def test_uppercase_on_assign(self):
        charset = Charset(allowed="")
        self.assertEqual("", charset.allowed)

        charset.allowed = "abc"
        self.assertEqual("ABC", charset.allowed)

    def test_replacements(self):
        charset = Charset(allowed="abc", replacements=[{"from": "x", "to": "a"}])
        cleaned = CharsetConverter(charset).clean(" abcx ")
        self.assertEqual("ABCA", cleaned)

    def test_replacement_to_non_allowed(self):
        charset = Charset(allowed="abc", replacements=[{"from": "x", "to": "y"}])
        cleaned = CharsetConverter(charset).clean(" abcx ")
        self.assertEqual("ABC", cleaned)

    def test_legacy_charset(self):
        cc = CharsetConverter(legacy_lpr_charset)
        clean = cc.clean("abcdefghijklmnopqrstuvwxyz1234567890üäö-_'*:;")
        self.assertEqual("ABCDEFGHIJKLMN0P0RSTUVWXYZ1234567890UA0", clean)



from datetime import datetime
from enum import Enum
from typing import List, Optional

from dateutil import rrule

from openmodule import config
from pydantic import validator
from openmodule.models.base import OpenModuleModel, ZMQMessage, Gateway


class Medium(str, Enum):
    LPR = "lpr"
    NFC = "nfc"
    PIN = "pin"
    QR = "qr"


class Category(str, Enum):
    BOOKED_DIGIMON = "booked-digimon"
    BOOKED_EMPLOYEE = "booked-employee"
    BOOKED_VISITOR = "booked-visitor"
    PERMANENT_DIGIMON = "permanent-digimon"
    PERMANENT_EMPLOYEE = "permanent-employee"
    FILLER_EMPLOYEE = "filler-employee"
    FILLER_DIGIMON = "filler-digimon"
    FILLER_VISITOR_BUTTON = "filler-visitor-button"
    FILLER_VISITOR_UNEXPECTED = "filler-visitor-unexpected"
    UNKNOWN_CATEGORY = "unknown-category"


class AccessRequest(OpenModuleModel):
    name: str
    gateway: Optional[Gateway] = None
    medium_type: Medium
    id: str


def check_recurrence(cls, recurrence, values, **kwargs):
    if recurrence == "":
        return recurrence
    if recurrence and not values.get("duration"):
        raise ValueError("set duration when using recurrence")
    try:
        rrule.rrulestr(recurrence)
    except Exception as e:
        raise ValueError(f"recurrence is not valid {e}")
    return recurrence


class Access(OpenModuleModel):
    category: Category
    start: datetime
    end: Optional[datetime] = None
    duration: Optional[int]
    recurrence: Optional[str]
    zone: Optional[str]
    occupant_check: bool = False
    infos: Optional[dict]
    user: str

    _check_recurrence = validator("recurrence", allow_reuse=True)(check_recurrence)

    def dict(self, **kwargs):
        data = super().dict(**kwargs)
        if "start" in data:
            data["start"] = data["start"].timestamp()
        if "end" in data:
            if data["end"] is not None:
                data["end"] = data["end"].timestamp()
        return data


class MediumAccesses(OpenModuleModel):
    accesses: List[Access]
    id: str
    type: str


class AccessResponse(OpenModuleModel):
    success: bool = False
    medium: MediumAccesses


class CountMessage(ZMQMessage):
    resource: str = config.resource()
    user: str
    gateway: Gateway
    medium_type: Medium
    id: str
    count: int
    transaction_id: str
    zone: str
    category: Category
    real: bool
    access_data: Optional[dict]
    error: Optional[str]
    previous_transaction_id: Optional[List[str]]  # double_entry, choose_random error
    previous_user: Optional[str]  # user_changed error
    previous_medium_type: Optional[str]  # medium_changed, medium_id_changed error
    previous_id: Optional[str]  # medium_changed, medium_id_changed error
    chosen: Optional[dict]  # choose_random error



from pydantic import validator, Field
from typing import List, Tuple, Union

from openmodule.models.base import OpenModuleModel


class Replacement(OpenModuleModel):
    c_from: str = Field(alias="from")
    c_to: str = Field(alias="to")

    @validator("c_from")
    def c_from_is_uppercase(cls, v):
        return v.upper()

    @validator("c_to")
    def c_to_is_uppercase(cls, v):
        return v.upper()


class Charset(OpenModuleModel):
    replacements: List[Replacement] = []
    allowed: str = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZÄÜÖ_"

    @validator("allowed")
    def allowed_is_uppercase(cls, v):
        return v.upper()


class CharsetConverter:
    def __init__(self, charset: Charset):
        self.replacements = charset.replacements
        self.allowed_chars = charset.allowed

    def _replace(self, text):
        for replacement in self.replacements:
            text = text.replace(replacement.c_from, replacement.c_to)
        return text

    def _remove_unknown(self, text):
        return "".join(x for x in text if x in self.allowed_chars)

    def clean(self, text):
        return self._remove_unknown(self._replace(text.upper()))


def _build_charset(allowed: str, replacements: Union[List, Tuple]) -> Charset:
    return Charset(
        allowed=allowed,
        replacements=({"from": f, "to": t} for f, t in replacements)
    )


legacy_lpr_charset = _build_charset(
    allowed="0123456789abcdefghijklmnprstuvwxyz",
    replacements=(("Ä", "A"), ("Ü", "U"), ("Ö", "O"), ("O", "0"), ("Q", "0"))
)

full_lpr_charset = _build_charset(
    allowed="0123456789ABCDEFGHIJKLMNOQPRSTUVWXYZÄÜÖ",
    replacements=((" ", ""), ("-", ""), (".", ""))
)



from enum import Enum

import orjson
import zmq
from datetime import datetime
from pydantic import Field, BaseModel
from pydantic.main import ROOT_KEY


def _donotuse(v, *, default):
    assert False, "please use json_bytes"


class OpenModuleModel(BaseModel):
    class Config:
        validate_assignment = True

        json_loads = orjson.loads
        json_dumps = _donotuse

    def json_bytes(self):
        data = self.dict()
        if self.__custom_root_type__:
            data = data[ROOT_KEY]
        return orjson.dumps(data)

    def dict(self, **kwargs):
        kwargs.setdefault("by_alias", True)
        kwargs.setdefault("exclude_none", True)
        return super().dict(**kwargs)


class EmptyModel(OpenModuleModel):
    pass


class ZMQMessage(OpenModuleModel):
    timestamp: datetime = Field(default_factory=datetime.now)
    name: str
    type: str

    def publish_on_topic(self, pub_socket: zmq.Socket, topic: bytes):
        pub_socket.send_multipart((topic, self.json_bytes()))

    def dict(self, **kwargs):
        data = super().dict(**kwargs)
        if "timestamp" in data:
            data["timestamp"] = data["timestamp"].timestamp()
        return data


class Direction(str, Enum):
    UNKNOWN = ""
    IN = "in"
    OUT = "out"


class Gateway(OpenModuleModel):
    gate: str = ""
    direction: Direction = ""

    def __str__(self):
        return f"{self.gate}/{self.direction}"



import logging
import threading
from collections import namedtuple
from typing import Callable, Any, Dict, Optional, Type
from uuid import UUID

import zmq
from pydantic import ValidationError, BaseModel
from pydantic.main import ROOT_KEY

from openmodule.messaging import get_sub_socket, get_pub_socket, receive_message_from_socket
from openmodule.models.rpc import RPCResponse, RPCRequest
from openmodule.rpc.common import channel_to_request_topic, channel_to_response_topic
from openmodule.threading import get_thread_wrapper

CallbackEntry = namedtuple("CallbackEntry", ["timestamp", "result"])
HandlerEntry = namedtuple("HandlerEntry", ["request_class", "response_class", "handler"])




def gateway_filter(gate=None, direction=None):
    def _filter(request, message, handler):
        gateway = request.get("gateway")
        if not gateway:
            return False
        return (not gate or gate == gateway.get("gate")) and \
               (not direction or direction == gateway.get("direction"))

    return _filter


class Filter:
    def __init__(self, filter, channel, type):
        self.filter = filter
        self.channel = channel
        self.type = type

    def check(self, request, message, handler, channel):
        if (self.channel is None or channel == self.channel) and (self.type is None or self.type == message.type):
            if self.filter(request=request, message=message, handler=handler):
                return True
            return False
        # filter does not apply
        return None


class RPCServer(object):
    def __init__(self, context, config):
        self.name = config.NAME
        self.sub = get_sub_socket(context, config)
        self.pub = get_pub_socket(context, config)
        self.handlers: Dict[tuple, HandlerEntry] = {}
        self.filters = []
        self.running = True
        self.log = logging.getLogger(self.__class__.__name__)
        self.thread = None

    def add_filter(self, filter: Callable[[], bool], channel=None, type=None):
        """
        :param filter: filter function with paramters reqeust, message, handler
        :param channel: specific channel to filter or all channels if None
        :param type: specific type to filter or all types if None
        """
        self.filters.append(Filter(filter=filter, channel=channel, type=type))

    def shutdown(self, timeout=3):
        self.running = False
        if self.thread:
            self.thread.join(timeout=timeout)

    def register_handler(self, channel: str, type: str, request_class: Type[BaseModel], response_class: Type[BaseModel],
                         handler: Callable):
        """
        :param channel: rpc channel you want to subscribe to, patterns/wildcards are not supported
        :param type: the request type
        :param handler: request handler of the form (request:dict, meta:dict) -> any
                        if the handler returns a dict which contains the key "status",
                        then the value of "status" will be used as the rpc response status.
                        There are no other reserved keys except "status".
        """
        if (channel, type) in self.handlers:
            raise ValueError(f"handler for {channel}:{type} is already registered")

        self.log.debug("register handler {}:{} -> {}".format(channel, type, handler))
        self.sub.subscribe(channel_to_request_topic(channel.encode("ascii")))
        self.handlers[(channel, type)] = HandlerEntry(request_class, response_class, handler)

    def _channel_from_topic(self, topic: bytes) -> str:
        return topic.split(b"-", 2)[-1].decode("ascii")

    def should_process_message(self, request, message, handler, channel):
        if self.filters:
            result = True
            for filter in self.filters:
                ok = filter.check(request=request, message=message, handler=handler, channel=channel)
                if ok:
                    return True
                # filter applies and is not ok
                elif ok is False:
                    result = False
            return result
        return True

    def find_handler(self, channel: str, type: str) -> HandlerEntry:
        return self.handlers.get((channel, type))

    def run_as_thread(self):
        assert not self.thread, "cannot run the same rpc server multiple times as thread"
        self.thread = threading.Thread(target=get_thread_wrapper(self.run))
        self.thread.start()
        return self.thread

    def process_rpc(self, channel, message: RPCRequest) -> Optional[Dict]:
        handler: HandlerEntry = self.find_handler(channel, message.type)
        if handler:
            try:
                request = handler.request_class(**message.request)
            except ValidationError as e:
                return {"status": "validation_error", "exception": e.json()}

            try:
                if not self.should_process_message(request, message, handler, channel):
                    return None
            except Exception as e:
                self.log.exception("exception in message filter. request not processed and error returned")
                return {"status": "filter_error", "exception": str(e)}

            try:
                response = handler.handler(request, message)
                if isinstance(response, dict):
                    return handler.response_class(**response).dict()
                elif response is None:
                    return {}  # None is used for when we do not handle the function, so we convert to dict here
                elif isinstance(response, handler.response_class):
                    data = response.dict()
                    if response.__custom_root_type__:
                        data = data[ROOT_KEY]
                    return data
                else:
                    raise Exception("rpc handler must return None, a dict or a instance of its response_class")
            except Exception as e:
                self.log.exception("exception in handler {}:{}".format(channel, type))
                return {"status": "handler_error", "exception": str(e)}
        else:
            self.log.warning("no handler found for {}:{}".format(channel, type))
            return None

    def run(self):
        poller = zmq.Poller()
        poller.register(self.sub, zmq.POLLIN)
        try:
            while self.running:
                socks = dict(poller.poll(timeout=1000))
                if socks.get(self.sub) != zmq.POLLIN:
                    continue
                topic, message = receive_message_from_socket(self.sub)
                if topic is None:
                    continue
                channel = self._channel_from_topic(topic)

                try:
                    message = RPCRequest(**message)
                    message_type = message.type
                    rpc_id = message.rpc_id
                except ValidationError as e:
                    message_type = "unknown"
                    rpc_id = None
                    response = {"status": "validation_error", "exception": e.json()}
                else:
                    response = self.process_rpc(channel, message)

                if response is not None:
                    self.log.debug("response: {}".format(response))
                    if "status" not in response:
                        response["status"] = "ok"

                    result = RPCResponse(
                        type=message_type,
                        rpc_id=rpc_id,
                        response=response,
                        name=self.name
                    )
                    result.publish_on_topic(
                        self.pub,
                        channel_to_response_topic(channel.encode("ascii")),
                    )
        except zmq.ContextTerminated:
            pass
        finally:
            self.sub.close()
            self.pub.close()



from enum import Enum
from typing import TYPE_CHECKING, Dict, Optional

from pydantic import validator

from openmodule.models.base import ZMQMessage

if TYPE_CHECKING:  # pragma: no cover
    from openmodule.core import OpenModuleCore


class AlertStatus(str, Enum):
    ok = "ok"
    error = "error"
    offline = "offline"


class AlertHandleType(str, Enum):
    state = "state"
    state_change = "state_change"
    count = "count"


class AlertMessage(ZMQMessage):
    type: str = "alert"
    status: AlertStatus
    alert_meta: Dict
    package: str
    alert_type: str
    source: Optional[str]
    handle_type: AlertHandleType
    value: Optional[float]

    @validator("value")
    def require_value_for_state_type(cls, v, values):
        if values["handle_type"] == AlertHandleType.state:
            assert v is not None, "value must not be None for alerts with handle_type='state'"
        return v



from typing import Optional, Tuple, Dict

import logging
import orjson
import zmq


def get_sub_socket(context, config, linger=0) -> zmq.Socket:
    socket = context.socket(zmq.SUB)
    socket.setsockopt(zmq.LINGER, linger)
    socket.connect(config.BROKER_PUB)
    return socket


def get_pub_socket(context, config, linger=100) -> zmq.Socket:
    socket = context.socket(zmq.PUB)
    socket.setsockopt(zmq.LINGER, linger)
    socket.connect(config.BROKER_SUB)
    return socket


def receive_message_from_socket(sub_socket: zmq.Socket) -> Tuple[Optional[bytes], Optional[Dict]]:
    """
    :param sub_socket: zmq socket to receive
    :return: returns None if an invalid message was received
             otherwise returns a tuple containing (topic: bytes, message: dict)
    """
    parts = sub_socket.recv_multipart()
    if len(parts) != 2:
        logging.error(f"received a zmq message with an invalid number of parts: {len(parts)}")
        return None, None

    topic, message = parts

    try:
        message_dict = orjson.loads(message)
    except orjson.JSONDecodeError as e:
        logging.error(f"received a zmq message with an invalid json. parsing error: {str(e)}")
        return None, None
    else:
        if not isinstance(message_dict, dict):
            logging.error(f"received a zmq message which was not a dict but a {type(message)}")
            return None, None
        else:
            return topic, message_dict
